# Generated by Django 2.1.8 on 2019-04-15 12:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("members", "0010_memberbalance"),
        ("mails", "0007_auto_20180929_0915"),
    ]

    operations = [
        migrations.AddField(
            model_name="email",
            name="balance",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="reminder_mails",
                to="members.MemberBalance",
            ),
        )
    ]



from contextlib import suppress

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.template.defaultfilters import register
from django.template.loader import TemplateDoesNotExist, get_template
from django.urls import reverse
from django.utils.html import escape
from django.utils.safestring import mark_safe

from byro.common.signals import log_formatters
from byro.documents.models import get_document_category_names

FORMATTER_REGISTRY = {}


def default_formatter(entry):
    with suppress(TemplateDoesNotExist):
        tmpl = get_template(f"log_entry/{entry.action_type}.html")
        return tmpl.render({"log_entry": entry})

    data = dict(entry.data or {})
    data.pop("source", None)

    co = entry.content_object
    related_object = ""
    if co:
        url = None
        if hasattr(co, "get_absolute_url"):
            url = co.get_absolute_url()
        elif isinstance(co, get_user_model()):
            url = reverse("office:settings.users.detail", kwargs={"pk": co.pk})

        if url:
            related_object = mark_safe(
                f' (<a href="{escape(url)}">{escape(str(co))}</a>)'
            )
        else:
            related_object = mark_safe(f" ({escape(co)})")

    extra_data = ""
    if data:
        extra_data = f" {escape(str(data))}"

    return mark_safe(f"{escape(str(entry.action_type))}{related_object}{extra_data}")


@register.filter(name="format_log_entry")
def format_log_entry(entry):
    if not FORMATTER_REGISTRY:
        for _receiver, response in log_formatters.send_robust(sender=__name__):
            if response and not isinstance(response, Exception):
                FORMATTER_REGISTRY.update(response)

    return FORMATTER_REGISTRY.get(entry.action_type, default_formatter)(entry)


@register.filter(name="format_log_source")
def format_log_source(entry):
    user = ""
    if entry.user:
        user = mark_safe(f'<span class="fa fa-user"></span> {escape(entry.user)}')

    source = entry.data.get("source", "")
    if source.startswith("internal: "):
        source = mark_safe(f'<span class="fa fa-gears"></span> {escape(source[10:])}')

    if entry.user:
        if entry.data.get("source", None) == str(entry.user):
            return user
        else:
            return mark_safe(f"{source} (via {user})")
    else:
        return source


@register.filter(name="format_log_object")
def format_log_object(obj, key=None):
    with suppress(Exception):
        if "object" in obj and "ref" in obj and "value" in obj:
            with suppress(Exception):
                content_object = ContentType.objects.get(
                    app_label=obj["ref"][0], model=obj["ref"][1]
                ).get_object_for_this_type(pk=obj["ref"][2])

                if obj["value"] == str(content_object):
                    url = content_object.get_absolute_url()
                    str_val = mark_safe(escape(str(obj["value"])))

                    if hasattr(content_object, "get_object_icon"):
                        icon = content_object.get_object_icon()
                    else:
                        icon = ""

                    if url:
                        return mark_safe(f'{icon}<a href="{escape(url)}">{str_val}</a>')

            return mark_safe(
                "<i>{} object</i>: {!r}".format(
                    escape(str(obj["object"])), escape(str(obj["value"]))
                )
            )

        if key == "category" and "." in obj:
            cats = get_document_category_names()
            if obj in cats:
                return f"{cats[obj]} ({obj})"

        if key == "content_hash":
            parts = str(obj).split(":", 1)
            if len(parts) == 2:
                return mark_safe(
                    '{}: <tt title="{}">{} &hellip; {}</tt>'.format(
                        escape(parts[0]),
                        escape(parts[1]),
                        escape(parts[1][: (6 * 2)]),
                        escape(parts[1][-(6 * 2) :]),
                    )
                )

    return obj


@register.filter(name="items_sorted")
def items_sorted(data):
    return sorted(data)



# Generated by Django 3.1 on 2020-08-18 20:02

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bookkeeping", "0016_auto_20181013_1611"),
    ]

    operations = [
        migrations.AlterField(
            model_name="booking",
            name="data",
            field=models.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="data",
            field=models.JSONField(null=True),
        ),
    ]



from typing import Tuple

from dateutil.relativedelta import relativedelta

from byro.members.models import Membership


def get_member_statistics_for_month(month, year) -> Tuple[int, int]:
    joins = Membership.objects.filter(start__month=month, start__year=year).count()
    quits = Membership.objects.filter(end__month=month, end__year=year).count()
    return joins, quits


def get_member_statistics():
    """Returns a list of tuples of the form ((year, month), joins, quits)."""
    first_member = Membership.objects.order_by("start").first()
    if not first_member:
        return []

    date = first_member.start
    end_member = Membership.objects.filter(end__isnull=False).order_by("-end").first()
    if not end_member:
        return []
    end = end_member.end
    result = []

    while date <= end:
        result.append(
            (
                (date.year, date.month),
                *get_member_statistics_for_month(date.month, date.year),
            )
        )
        date += relativedelta(months=1)
    return result



# Generated by Django 1.11.4 on 2017-09-19 18:32

from django.db import migrations


def create_accounts(apps, schema_editor):
    from byro.bookkeeping.models import AccountCategory

    Account = apps.get_model("bookkeeping", "Account")
    for category in (
        "member_donation",
        "member_fees",
    ):  # Not the constants, because the class got refactored
        Account.objects.create(account_category=category)


def delete_accounts(apps, schema_editor):
    Account = apps.get_model("bookkeeping", "Account")
    Account.objects.all().delete()


class Migration(migrations.Migration):
    dependencies = [("bookkeeping", "0004_auto_20170919_1832")]

    operations = [migrations.RunPython(create_accounts, delete_accounts)]



# Generated by Django 1.11.6 on 2017-10-12 19:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("common", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="configuration",
            name="mail_from",
            field=models.EmailField(
                blank=True,
                max_length=100,
                null=True,
                verbose_name="e-mail sender address",
            ),
        )
    ]



from copy import deepcopy

from django.db import models, transaction
from django.db.models import Q
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.utils.translation import override
from i18nfield.fields import I18nCharField, I18nTextField

from byro.common.models.auditable import Auditable
from byro.mails.send import SendMailException
from byro.members.models import Member


class MailTemplate(Auditable, models.Model):
    subject = I18nCharField(max_length=200, verbose_name=_("Subject"))
    text = I18nTextField(verbose_name=_("Text"))
    reply_to = models.EmailField(
        max_length=200,
        blank=True,
        null=True,
        verbose_name=_("Reply-To"),
        help_text=_(
            "Change the Reply-To address if you do not want to use the default orga address"
        ),
    )
    bcc = models.CharField(
        max_length=1000,
        blank=True,
        null=True,
        verbose_name=_("BCC"),
        help_text=_(
            "Enter comma separated addresses. Will receive a blind copy of every mail sent from this template. This may be a LOT!"
        ),
    )

    def __str__(self):
        return f"{self.subject}"

    def to_mail(
        self,
        email,
        locale=None,
        context=None,
        skip_queue=False,
        attachments=None,
        save=True,
    ):
        from byro.common.models import Configuration

        config = Configuration.get_solo()
        locale = locale or config.language
        with override(locale):
            context = context or dict()
            try:
                subject = str(self.subject).format(**context)
                text = str(self.text).format(**context)
            except KeyError as e:
                raise SendMailException(
                    f"Experienced KeyError when rendering Text: {e}"
                )

            mail = EMail(
                to=email,
                reply_to=self.reply_to,
                bcc=self.bcc,
                subject=subject,
                text=text,
                template=self,
            )
            if save:
                mail.save()
                if attachments:
                    for a in attachments:
                        mail.attachments.add(a)
                if skip_queue:
                    mail.send()
        return mail

    def get_absolute_url(self):
        return reverse("office:mails.templates.view", kwargs={"pk": self.pk})

    def get_object_icon(self):
        return mark_safe('<i class="fa fa-envelope-o"></i> ')


class EMail(Auditable, models.Model):
    to = models.CharField(
        max_length=1000,
        verbose_name=_("To"),
        help_text=_("One email address or several addresses separated by commas."),
    )
    reply_to = models.CharField(
        max_length=1000, null=True, blank=True, verbose_name=_("Reply-To")
    )
    cc = models.CharField(
        max_length=1000,
        null=True,
        blank=True,
        verbose_name=_("CC"),
        help_text=_("One email address or several addresses separated by commas."),
    )
    bcc = models.CharField(
        max_length=1000,
        null=True,
        blank=True,
        verbose_name=_("BCC"),
        help_text=_("One email address or several addresses separated by commas."),
    )
    subject = models.CharField(max_length=200, verbose_name=_("Subject"))
    members = models.ManyToManyField(to="members.Member", related_name="emails")
    text = models.TextField(verbose_name=_("Text"))
    sent = models.DateTimeField(null=True, blank=True, verbose_name=_("Sent at"))
    template = models.ForeignKey(
        to=MailTemplate, null=True, blank=True, on_delete=models.SET_NULL
    )
    attachments = models.ManyToManyField(to="documents.Document", related_name="mails")
    balance = models.ForeignKey(
        to="members.MemberBalance",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        related_name="reminder_mails",
    )

    @property
    def attachment_ids(self):
        if hasattr(self, "attachments"):
            return list(self.attachments.all().values_list("pk", flat=True))
        return []

    def process_special_to(self):
        # Handles "special:" adresses, except "special:all" which is handled inside send()
        if self.to.startswith("special:"):
            if self.to.startswith("special:member:"):
                member = Member.all_objects.get(pk=self.to.split(":", 2)[2])
                self.to = member.email
                self.members.add(member)
                self.save(update_fields=["to"])

    @transaction.atomic
    def send(self):
        if self.sent:
            raise TypeError("This mail has been sent already. It cannot be sent again.")

        self.process_special_to()

        from byro.common.models import Configuration

        config = Configuration.get_solo()

        if self.to == "special:all" or not self.to.startswith("special:"):
            send_tos = []

            if self.to == "special:all":
                for member in (
                    Member.objects.filter(
                        Q(memberships__start__lte=now().date())
                        & (
                            Q(memberships__end__isnull=True)
                            | Q(memberships__end__gte=now().date())
                        )
                    )
                    .filter(email__isnull=False)
                    .exclude(email="")
                    .all()
                ):
                    send_tos.append(member)
                    self.members.add(member)

            else:
                to_addrs = self.to.split(",")
                for addr in to_addrs:
                    member = Member.all_objects.filter(
                        email__iexact=addr.lower()
                    ).first()
                    if member:
                        send_tos.append(member)
                        self.members.add(member)
                    else:
                        send_tos.append(addr)

            headers = {}
            if self.reply_to:
                headers["Reply-To"] = self.reply_to

            from byro.mails.send import mail_send_task

            for addr in send_tos:
                body = self.text
                if isinstance(addr, Member):
                    signature = _(
                        "You are receiving this email due to your membership in {name}."
                    ).format(name=config.name)
                    signature += "\n"
                    signature += _(
                        "You can see your member page at this URL: {url}"
                    ).format(url=addr.profile_memberpage.get_url())
                    body += "\n\n-- \n" + signature
                    addr = addr.email
                mail_send_task(
                    to=[addr],
                    subject=self.subject,
                    body=body,
                    sender=config.mail_from,
                    cc=(self.cc or "").split(","),
                    bcc=(self.bcc or "").split(","),
                    attachments=self.attachment_ids,
                    headers=headers,
                )

        self.sent = now()
        self.save(update_fields=["sent"])

    def copy_to_draft(self):
        new_mail = deepcopy(self)
        new_mail.pk = None
        new_mail.sent = None
        new_mail.save()
        return new_mail



# Generated by Django 3.2.10 on 2021-12-08 20:21

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("common", "0018_auto_20200820_2018"),
    ]

    operations = [
        migrations.AddField(
            model_name="configuration",
            name="currency_postfix",
            field=models.BooleanField(
                default=True,
                help_text="Controls whether the currency symbol comes before or after the monetary value",
                verbose_name="Show currency symbol after value",
            ),
        ),
        migrations.AddField(
            model_name="configuration",
            name="currency_symbol",
            field=models.CharField(
                default="€",
                help_text="E.g. €",
                max_length=8,
                verbose_name="Currency symbol",
            ),
        ),
        migrations.AddField(
            model_name="configuration",
            name="display_cents",
            field=models.BooleanField(
                default=True,
                help_text="When enabled, monetary values include two decimal fractional digits",
                verbose_name="Display cents",
            ),
        ),
    ]



from django import forms
from django.contrib.auth.models import User
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, FormView, ListView, UpdateView

from byro.common.models import LogEntry


class UserForm(forms.ModelForm):
    password = forms.CharField(label=_("Password"), widget=forms.PasswordInput)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["last_name"].label = _("Name")

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        password = self.cleaned_data.get("password")
        if password:
            self.instance.set_password(password)
            self.instance.save()

    class Meta:
        model = User
        fields = [
            "username",
            "last_name",
            "email",
            "is_superuser",
            "is_staff",
        ]


class UserListView(ListView):
    template_name = "office/user/list.html"
    context_object_name = "users"
    model = User
    paginate_by = 25


class UserCreateView(FormView):
    template_name = "office/user/add.html"
    model = User
    form_class = UserForm

    def form_valid(self, form):
        form.save()
        self.form = form
        LogEntry.objects.create(
            content_object=form.instance,
            user=self.request.user,
            action_type="byro.common.user.created",
        )
        return super().form_valid(form)

    def get_success_url(self):
        return reverse(
            "office:settings.users.detail", kwargs={"pk": self.form.instance.pk}
        )


class UserDetailView(UpdateView):
    template_name = "office/user/detail.html"
    context_object_name = "user"
    model = User
    form_class = UserForm

    def form_valid(self, form):
        LogEntry.objects.create(
            content_object=form.instance,
            user=self.request.user,
            action_type="byro.common.user.updated",
        )
        return super().form_valid(form)

    def get_object(self):
        return User.objects.get(pk=self.kwargs["pk"])

    def get_success_url(self):
        return reverse("office:settings.users.detail", kwargs={"pk": self.kwargs["pk"]})


# FIXME No implemented yet
class UserDeleteView(DetailView):
    model = User
    context_object_name = "user"



# Generated by Django 1.11.4 on 2017-09-19 18:40

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("bookkeeping", "0005_auto_20170919_1832")]

    operations = [migrations.RemoveField(model_name="account", name="member")]



from django.urls import include, re_path

from .views import (
    accounts,
    dashboard,
    documents,
    mails,
    members,
    settings,
    transactions,
    upload,
    users,
)

app_name = "office"
urlpatterns = [
    re_path(
        "^settings/initial$",
        settings.InitialSettings.as_view(),
        name="settings.initial",
    ),
    re_path("^settings/log$", settings.LogView.as_view(), name="settings.log"),
    re_path(
        "^settings/about$", settings.AboutByroView.as_view(), name="settings.about"
    ),
    re_path(
        "^settings/registration$",
        settings.RegistrationConfigView.as_view(),
        name="settings.registration",
    ),
    re_path(
        "^settings/users/$", users.UserListView.as_view(), name="settings.users.list"
    ),
    re_path(
        "^settings/users/add$",
        users.UserCreateView.as_view(),
        name="settings.users.add",
    ),
    re_path(
        r"^settings/users/(?P<pk>\d+)/$",
        users.UserDetailView.as_view(),
        name="settings.users.detail",
    ),
    re_path("^settings$", settings.ConfigurationView.as_view(), name="settings.base"),
    re_path("^$", dashboard.DashboardView.as_view(), name="dashboard"),
    re_path(
        r"^members/typeahead$",
        members.MemberListTypeaheadView.as_view(),
        name="members.typeahead",
    ),
    re_path(r"^members/list$", members.MemberListView.as_view(), name="members.list"),
    re_path(
        r"^members/list/export$",
        members.MemberListExportView.as_view(),
        name="members.list.export",
    ),
    re_path(
        r"^members/list/import$",
        members.MemberListImportView.as_view(),
        name="members.list.import",
    ),
    re_path(
        r"^members/list/disclosure$",
        members.MemberDisclosureView.as_view(),
        name="members.disclosure",
    ),
    re_path(
        r"^members/list/balance$",
        members.MemberBalanceView.as_view(),
        name="members.balance",
    ),
    re_path(r"^members/add$", members.MemberCreateView.as_view(), name="members.add"),
    re_path(
        r"^members/view/(?P<pk>\d+)/",
        include(
            [
                re_path("data$", members.MemberDataView.as_view(), name="members.data"),
                re_path(
                    "timeline$",
                    members.MemberTimelineView.as_view(),
                    name="members.timeline",
                ),
                re_path(
                    "finance$",
                    members.MemberFinanceView.as_view(),
                    name="members.finance",
                ),
                re_path(
                    "operations$",
                    members.MemberOperationsView.as_view(),
                    name="members.operations",
                ),
                re_path(
                    "record-disclosure$",
                    members.MemberRecordDisclosureView.as_view(),
                    name="members.record-disclosure",
                ),
                re_path("log$", members.MemberLogView.as_view(), name="members.log"),
                re_path(
                    "mails$", members.MemberMailsView.as_view(), name="members.mails"
                ),
                re_path(
                    "documents$",
                    members.MemberDocumentsView.as_view(),
                    name="members.documents",
                ),
                re_path(
                    "$", members.MemberDashboardView.as_view(), name="members.dashboard"
                ),
            ]
        ),
    ),
    re_path(
        r"^transactions/(?P<pk>\d+)/",
        transactions.TransactionDetailView.as_view(),
        name="finance.transactions.detail",
    ),
    re_path(
        "^upload/list", upload.UploadListView.as_view(), name="finance.uploads.list"
    ),
    re_path(
        r"^upload/process/(?P<pk>\d+)",
        upload.UploadProcessView.as_view(),
        name="finance.uploads.process",
    ),
    re_path(
        r"^upload/match/(?P<pk>\d+)",
        upload.UploadMatchView.as_view(),
        name="finance.uploads.match",
    ),
    re_path("^upload/add", upload.CsvUploadView.as_view(), name="finance.uploads.add"),
    re_path(
        "^documents/add", documents.DocumentUploadView.as_view(), name="documents.add"
    ),
    re_path(
        r"^documents/(?P<pk>\d+)/(?P<filename>[^/]+)",
        documents.DocumentDownloadView.as_view(),
        name="documents.download",
    ),
    re_path(
        r"^documents/(?P<pk>\d+)",
        documents.DocumentDetailView.as_view(),
        name="documents.detail",
    ),
    re_path(
        "^accounts/$", accounts.AccountListView.as_view(), name="finance.accounts.list"
    ),
    re_path(
        "^accounts/add$",
        accounts.AccountCreateView.as_view(),
        name="finance.accounts.add",
    ),
    re_path(
        r"^accounts/(?P<pk>\d+)/$",
        accounts.AccountDetailView.as_view(),
        name="finance.accounts.detail",
    ),
    re_path(
        r"^accounts/(?P<pk>\d+)/delete$",
        accounts.AccountDeleteView.as_view(),
        name="finance.accounts.delete",
    ),
    re_path(
        "^mails/(?P<pk>[0-9]+)$", mails.MailDetail.as_view(), name="mails.mail.view"
    ),
    re_path(
        "^mails/(?P<pk>[0-9]+)/copy$", mails.MailCopy.as_view(), name="mails.mail.copy"
    ),
    re_path(
        "^mails/(?P<pk>[0-9]+)/delete$",
        mails.OutboxPurge.as_view(),
        name="mails.mail.delete",
    ),
    re_path(
        "^mails/(?P<pk>[0-9]+)/send$",
        mails.OutboxSend.as_view(),
        name="mails.mail.send",
    ),
    re_path("^mails/compose$", mails.Compose.as_view(), name="mails.compose"),
    re_path("^mails/sent$", mails.SentMail.as_view(), name="mails.sent"),
    re_path("^mails/outbox$", mails.OutboxList.as_view(), name="mails.outbox.list"),
    re_path(
        "^mails/outbox/send$", mails.OutboxSend.as_view(), name="mails.outbox.send"
    ),
    re_path(
        "^mails/outbox/purge$", mails.OutboxPurge.as_view(), name="mails.outbox.purge"
    ),
    re_path(
        "^mails/templates$", mails.TemplateList.as_view(), name="mails.templates.list"
    ),
    re_path(
        "^mails/templates/add$",
        mails.TemplateCreate.as_view(),
        name="mails.templates.add",
    ),
    re_path(
        "^mails/templates/(?P<pk>[0-9]+)$",
        mails.TemplateDetail.as_view(),
        name="mails.templates.view",
    ),
    re_path(
        "^templates/(?P<pk>[0-9]+)/delete$",
        mails.TemplateDelete.as_view(),
        name="mails.templates.delete",
    ),
]



# Generated by Django 2.1.7 on 2019-02-26 11:16

import annoying.fields
import byro.public.models
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [("members", "0009_auto_20180512_1810")]

    operations = [
        migrations.CreateModel(
            name="MemberpageProfile",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "secret_token",
                    models.CharField(
                        blank=True,
                        default=byro.public.models.generate_default_token,
                        max_length=128,
                        null=True,
                    ),
                ),
                (
                    "is_visible_to_members",
                    models.BooleanField(
                        default=False, verbose_name="Consent: Visible to other members"
                    ),
                ),
                (
                    "member",
                    annoying.fields.AutoOneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="profile_memberpage",
                        to="members.Member",
                    ),
                ),
            ],
        )
    ]



from collections import OrderedDict

from django.utils.functional import classproperty


class ChoicesMeta(type):
    @classmethod
    def __prepare__(self, name, bases):
        return OrderedDict()

    def __new__(cls, name, parents, dct):
        if "valid_choices" not in dct:
            dct["valid_choices"] = [
                dct[key] for key in dct if isinstance(key, str) and key.upper() == key
            ]
        return super().__new__(cls, name, parents, dct)


class Choices(metaclass=ChoicesMeta):
    """Helper class to make choices available as class variables, expose a list
    with valid choices and at the same time generate the choices tuples for the
    model class.

    Usage:
        class MyChoices(Choices):
            ONE = 'one'
            TWO = 'dwa'

        class MyModel(models.Model):
            variant = models.CharField(
                max_length=MyChoices.max_length,
                choices=MyChoices.choices,
            )
    """

    @classproperty
    def choices(cls):
        return ((val, val) for val in cls.valid_choices)

    @classproperty
    def max_length(cls):
        if hasattr(cls, "valid_choices"):
            return max(len(val) for val in cls.valid_choices)
        else:
            return max(len(val) for val, _ in cls.choices)



from hashlib import sha512

import magic
from django.apps import apps
from django.db import models, transaction
from django.db.models.signals import pre_delete
from django.dispatch import receiver
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from byro.common.models import LogTargetMixin
from byro.common.models.choices import Choices


class DocumentDirection(Choices):
    INCOMING = "incoming"
    OUTGOING = "outgoing"
    OTHER = "other"


class Document(models.Model, LogTargetMixin):
    LOG_TARGET_BASE = "byro.documents.document"

    class Meta:
        ordering = ("-date", "title", "-id")

    document = models.FileField(upload_to="documents/%Y/%m/", max_length=1000)
    date = models.DateField(null=True, default=now)
    title = models.CharField(max_length=300, null=True)
    category = models.CharField(max_length=300, null=True)
    direction = models.CharField(
        choices=DocumentDirection.choices,
        max_length=DocumentDirection.max_length,
        default=DocumentDirection.OUTGOING,
    )
    member = models.ForeignKey(
        to="members.Member",
        related_name="documents",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    content_hash = models.CharField(max_length=300, null=True)

    template = _(
        """
Hi, {name},

Please find attached a document we wanted to send you/that you requested.

Thank you,
{association}
"""
    ).strip()

    def _get_log_properties(self):
        return {
            f.name: getattr(self, f.name)
            for f in self._meta.get_fields()
            if f.name not in ("id", "mails") and hasattr(self, f.name)
        }

    @cached_property
    def content_hash_ok(self):
        h = sha512()
        with self.document.open(mode="rb") as f:
            for chunk in f.chunks():
                h.update(chunk)
        return self.content_hash == f"sha512:{h.hexdigest()}"

    @cached_property
    def mime_type_guessed(self):
        with self.document.open(mode="rb") as f:
            chunk = next(f.chunks())
            return magic.from_buffer(chunk, mime=True)

    @transaction.atomic
    def save(self, *args, **kwargs):
        retval = super().save(*args, **kwargs)

        # Only store the hash the first *time* that a file is added
        # Unfortunately we cannot do that on the first *save*, because
        # the API allows adding file content after the fact
        if self.document and not self.content_hash:
            h = sha512()
            with self.document.open(mode="rb") as f:
                for chunk in f.chunks():
                    h.update(chunk)
            self.content_hash = f"sha512:{h.hexdigest()}"
            super().save(update_fields=["content_hash"])
            self.log(
                "internal: automatic checkpoint",
                ".stored",
                **self._get_log_properties(),
            )

        return retval

    def send(self, immediately=False, text=None, subject=None, email=None):
        from byro.common.models import Configuration
        from byro.mails.models import EMail

        us = Configuration.get_solo().name
        mail = EMail.objects.create(
            to=email or self.member.email,
            text=text
            or self.template.format(
                name=self.member.name if self.member else email, association=us
            ),
            subject=_("[{association}] Your document").format(association=us),
        )
        mail.attachments.add(self)
        mail.save()
        if immediately:
            mail.send()
        return mail

    def get_display(self):
        return "{} [{}] Document: {}".format(
            self.get_direction_display().capitalize(), self.category, self.title
        )

    def get_absolute_url(self):
        return reverse("office:documents.detail", kwargs={"pk": self.pk})

    def get_object_icon(self):
        return mark_safe('<i class="fa fa-file-o"></i> ')

    @property
    def basename(self):
        return self.document.name.split("/")[-1]


@receiver(pre_delete, sender=Document, dispatch_uid="documents_models__log_deletion")
def log_deletion(sender, instance, using, **kwargs):
    instance.log(
        "internal: automatic checkpoint", ".deleted", **instance._get_log_properties()
    )


def get_document_category_names():
    categories = {}

    for app in apps.get_app_configs():
        if hasattr(app, "ByroPluginMeta"):
            categories.update(getattr(app.ByroPluginMeta, "document_categories", {}))

    return categories



from datetime import timedelta

from dateutil.relativedelta import relativedelta
from django.core.management.base import BaseCommand
from django.db import transaction
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from byro.bookkeeping.models import Transaction
from byro.bookkeeping.special_accounts import SpecialAccounts
from byro.common.models.configuration import Configuration
from byro.members.models import FeeIntervals, Member, Membership, MembershipType

SOURCE_TEST_DATA = "Import of test data"


def make_date(delta, end=False):
    date = (now() - delta).date()
    date = date.replace(day=1)
    if end:
        date = date + relativedelta(month=1)
        return date - timedelta(days=1)
    return date


class Command(BaseCommand):
    help = "Introduce test data, including members and payments"
    leave_locale_alone = True

    def create_configs(self):
        config = Configuration.get_solo()
        config.name = "Der Verein e.V."
        config.address = "Erich-Weinert-Straße 53\n39104 Magdeburg"
        config.url = "https://dervereindervere.in"
        config.language = "de"
        config.currency = "EUR"
        config.mail_from = "verein@dervereindervere.in"
        config.backoffice_mail = "vorstanz@dervereindervere.in"
        config.registration_form = [
            {"name": "member__number", "position": 1},
            {"name": "member__name", "position": 2},
            {"name": "member__address", "position": 3},
            {"name": "member__email", "position": 4},
            {
                "name": "membership__start",
                "position": 5,
                "default_date": "beginning_month",
            },
            {"name": "membership__interval", "default": "1", "position": 6},
            {"name": "membership__amount", "default": "23", "position": 7},
        ]
        config.save()
        config.log(SOURCE_TEST_DATA, ".changed")

    def make_paid(self, member, vaguely=False, overly=False, donates=0, pays_for=None):
        member.update_liabilites()
        for index, liability in enumerate(
            member.bookings.filter(
                debit_account=SpecialAccounts.fees_receivable,
                transaction__value_datetime__lte=now(),
            ).all()
        ):
            if vaguely and index % 2 == 0:
                continue

            pure_amount = liability.amount if not overly else liability.amount * 2

            text = _("Member fee for {number}").format(number=member.number)

            if pays_for:
                amount = pure_amount * 2
                text += ", " + _("and for {number}").format(number=pays_for.number)
            else:
                amount = pure_amount

            if donates:
                amount += donates
                text += ", " + _("EUR {amount} donation").format(amount=donates)

            t = Transaction.objects.create(
                value_datetime=liability.transaction.value_datetime,
                user_or_context=SOURCE_TEST_DATA,
            )
            t.debit(
                memo=text,
                account=SpecialAccounts.bank,
                amount=amount,
                user_or_context=SOURCE_TEST_DATA,
            )
            if donates:
                t.credit(
                    account=SpecialAccounts.donations,
                    member=member,
                    amount=donates,
                    user_or_context=SOURCE_TEST_DATA,
                )
            t.credit(
                account=SpecialAccounts.fees_receivable,
                member=member,
                amount=pure_amount,
                user_or_context=SOURCE_TEST_DATA,
            )
            if pays_for:
                t.credit(
                    account=SpecialAccounts.fees_receivable,
                    member=pays_for,
                    amount=pure_amount,
                    user_or_context=SOURCE_TEST_DATA,
                )
            t.save()

    def create_membership_types(self):
        MembershipType.objects.create(name="Standard membership", amount=120)

    def create_members(self):
        has_left = Member.objects.create(
            number="1",
            name="Francis Foundingmember",
            address="Foo St 1\nSome Place",
            email="francis@group.org",
        )
        Membership.objects.create(
            member=has_left,
            start=make_date(relativedelta(years=2)),
            end=make_date(relativedelta(years=1), end=True),
            interval=FeeIntervals.MONTHLY,
            amount=10,
        )
        has_left.log(SOURCE_TEST_DATA, ".created")
        self.make_paid(has_left)

        does_not_pay = Member.objects.create(
            number="2",
            name="Yohnny Yolo",
            address="Bar St 1\nSome Distant Place",
            email="yolo@group.org",
        )
        Membership.objects.create(
            member=does_not_pay,
            start=make_date(relativedelta(years=4)),
            interval=FeeIntervals.MONTHLY,
            amount=10,
        )
        does_not_pay.log(SOURCE_TEST_DATA, ".created")
        does_not_pay.update_liabilites()

        pays_occasionally = Member.objects.create(
            number="3",
            name="Olga Occasional",
            address="Currently unknown",
            email="olga@group.org",
        )
        Membership.objects.create(
            member=pays_occasionally,
            start=make_date(relativedelta(years=4, months=6)),
            interval=FeeIntervals.MONTHLY,
            amount=10,
        )
        pays_occasionally.log(SOURCE_TEST_DATA, ".created")
        self.make_paid(pays_occasionally, vaguely=True)

        pays_regularly = Member.objects.create(
            number="4",
            name="Dennis Diligent",
            address="Best St 3\nFoo Town\nMy Country\nEarth\nUniverse",
            email="dennis@group.org",
        )
        Membership.objects.create(
            member=pays_regularly,
            start=make_date(relativedelta(years=1, months=3)),
            interval=FeeIntervals.MONTHLY,
            amount=10,
        )
        pays_regularly.log(SOURCE_TEST_DATA, ".created")
        self.make_paid(pays_regularly)

        pays_too_much = Member.objects.create(
            number="5",
            name="Omar Overachiever",
            address="SuperBest St 3\nSuperFoo Town",
            email="omar@group.org",
        )
        Membership.objects.create(
            member=pays_too_much,
            start=make_date(relativedelta(years=1)),
            interval=FeeIntervals.MONTHLY,
            amount=10,
        )
        pays_too_much.log(SOURCE_TEST_DATA, ".created")
        self.make_paid(pays_too_much, overly=True)

        will_join = Member.objects.create(
            number="6",
            name="Francine Futuremember",
            address="Future St 3\nFuture Town",
            email="francine@group.org",
        )
        Membership.objects.create(
            member=will_join,
            start=make_date(relativedelta(months=-2)),
            interval=FeeIntervals.MONTHLY,
            amount=10,
        )
        will_join.log(SOURCE_TEST_DATA, ".created")

        giver = Member.objects.create(
            number="7",
            name="George Giver",
            address="Generous St 3\nEnd-of-the-rainbow Hearth",
            email="george@group.org",
        )
        Membership.objects.create(
            member=giver,
            start=make_date(relativedelta(years=1)),
            interval=FeeIntervals.MONTHLY,
            amount=10,
        )
        giver.log(SOURCE_TEST_DATA, ".created")
        self.make_paid(giver, donates=5)

        is_payed_for = Member.objects.create(
            number="8",
            name="Peter Partner",
            address="Commune St 3\nFamily Shire",
            email="peter@group.org",
        )
        Membership.objects.create(
            member=is_payed_for,
            start=make_date(relativedelta(months=3)),
            interval=FeeIntervals.MONTHLY,
            amount=10,
        )
        is_payed_for.log(SOURCE_TEST_DATA, ".created")
        is_payed_for.update_liabilites()

        pays_other = Member.objects.create(
            number="9",
            name="Aaron Alsopayer",
            address="Commune St 3\nFamily Shire",
            email="aaron@group.org",
        )
        Membership.objects.create(
            member=pays_other,
            start=make_date(relativedelta(months=3)),
            interval=FeeIntervals.MONTHLY,
            amount=10,
        )
        pays_other.log(SOURCE_TEST_DATA, ".created")
        self.make_paid(pays_other, pays_for=is_payed_for)
        for member in Member.objects.all():
            member.update_name_fields()

    def create_bank_chaff(self):
        """Create some dummy traffic, and a couple of unmatched transactions on
        the bank account."""
        bank_account = SpecialAccounts.bank

        t = Transaction.objects.create(
            value_datetime=(now() - relativedelta(days=23)).date(),
            user_or_context=SOURCE_TEST_DATA,
        )
        t.debit(
            memo=_("Belated member fee payment for Olga"),
            account=bank_account,
            amount=20,
            user_or_context=SOURCE_TEST_DATA,
        )
        t.save()

        t = Transaction.objects.create(
            value_datetime=(now() - relativedelta(days=17)).date(),
            user_or_context=SOURCE_TEST_DATA,
        )
        t.debit(
            memo=_("George lives to give, donation"),
            account=bank_account,
            amount=42.23,
            user_or_context=SOURCE_TEST_DATA,
        )
        t.save()

        for month in range(1, 4):
            t = Transaction.objects.create(
                value_datetime=(now() - relativedelta(months=month)).date(),
                user_or_context=SOURCE_TEST_DATA,
            )
            t.credit(
                memo=_("Bank fees"),
                account=bank_account,
                amount=9.95,
                user_or_context=SOURCE_TEST_DATA,
            )
            t.save()

        t = Transaction.objects.create(
            value_datetime=(now() - relativedelta(days=21)).date(),
            user_or_context=SOURCE_TEST_DATA,
        )
        t.credit(
            memo=_("ACME Inc. thanks you for your patronage, sale of one halo kite"),
            account=bank_account,
            amount=123,
            user_or_context=SOURCE_TEST_DATA,
        )
        t.save()

        t = Transaction.objects.create(
            value_datetime=(now() - relativedelta(days=20)).date(),
            user_or_context=SOURCE_TEST_DATA,
        )
        t.credit(
            memo=_(
                "ACME Inc. thanks you for your patronage, sale of one emergency medkit"
            ),
            account=bank_account,
            amount=666,
            user_or_context=SOURCE_TEST_DATA,
        )
        t.save()

    @transaction.atomic()
    def handle(self, *args, **options):
        self.create_configs()
        self.create_membership_types()
        self.create_members()
        self.create_bank_chaff()



# Generated by Django 1.11.8 on 2018-01-11 18:07

import byro.common.models.auditable
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [("bookkeeping", "0007_auto_20171206_1919")]

    operations = [
        migrations.CreateModel(
            name="RealTransactionSource",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("source_file", models.FileField(upload_to="transaction_uploads/")),
            ],
            bases=(byro.common.models.auditable.Auditable, models.Model),
        ),
        migrations.AddField(
            model_name="realtransaction",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="transactions",
                to="bookkeeping.RealTransactionSource",
            ),
        ),
    ]



# Generated by Django 1.11.6 on 2017-10-12 19:30

import byro.common.models.auditable
from django.db import migrations, models
import i18nfield.fields


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="EMail",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "to",
                    models.CharField(
                        help_text="One email address or several addresses separated by commas.",
                        max_length=1000,
                        verbose_name="To",
                    ),
                ),
                (
                    "reply_to",
                    models.CharField(
                        blank=True, max_length=1000, null=True, verbose_name="Reply-To"
                    ),
                ),
                (
                    "cc",
                    models.CharField(
                        blank=True,
                        help_text="One email address or several addresses separated by commas.",
                        max_length=1000,
                        null=True,
                        verbose_name="CC",
                    ),
                ),
                (
                    "bcc",
                    models.CharField(
                        blank=True,
                        help_text="One email address or several addresses separated by commas.",
                        max_length=1000,
                        null=True,
                        verbose_name="BCC",
                    ),
                ),
                ("subject", models.CharField(max_length=200, verbose_name="Subject")),
                ("text", models.TextField(verbose_name="Text")),
                (
                    "sent",
                    models.DateTimeField(blank=True, null=True, verbose_name="Sent at"),
                ),
            ],
            bases=(byro.common.models.auditable.Auditable, models.Model),
        ),
        migrations.CreateModel(
            name="MailTemplate",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "subject",
                    i18nfield.fields.I18nCharField(
                        max_length=200, verbose_name="Subject"
                    ),
                ),
                ("text", i18nfield.fields.I18nTextField(verbose_name="Text")),
                (
                    "bcc",
                    models.CharField(
                        blank=True,
                        help_text="Enter comma separated addresses. Will receive a blind copy of every mail sent from this template. This may be a LOT!",
                        max_length=1000,
                        null=True,
                        verbose_name="BCC",
                    ),
                ),
            ],
            bases=(byro.common.models.auditable.Auditable, models.Model),
        ),
    ]



import collections
import collections.abc
import csv
import hashlib
from datetime import datetime, time
from decimal import Decimal
from functools import partial
from itertools import chain

import dateparser
import unicodecsv
from chardet import UniversalDetector
from dateutil.relativedelta import relativedelta
from django import forms
from django.conf import settings
from django.contrib import messages
from django.db import transaction
from django.db.models import Q
from django.dispatch import receiver
from django.http import JsonResponse, StreamingHttpResponse
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, FormView, ListView, TemplateView, View
from django.views.generic.list import (
    MultipleObjectMixin,
    MultipleObjectTemplateResponseMixin,
)

from byro.bookkeeping.models import Booking
from byro.bookkeeping.special_accounts import SpecialAccounts
from byro.common.models import Configuration, LogEntry
from byro.mails.models import EMail
from byro.members.forms import CreateMemberForm
from byro.members.models import Member, Membership
from byro.members.signals import (
    leave_member,
    leave_member_mail_information,
    leave_member_office_mail_information,
    new_member,
    new_member_mail_information,
    new_member_office_mail_information,
    update_member,
)
from byro.members.timeline import (
    augment_timeline,
    get_document_timeline,
    get_finance_timeline,
    get_mail_timeline,
    get_ops_timeline,
    sorted_merge,
)
from byro.office.signals import (
    member_dashboard_tile,
    member_list_importers,
    member_view,
)

from .documents import DocumentUploadForm


class MemberView(DetailView):
    context_object_name = "member"
    model = Member

    def get_member(self):
        return Member.all_objects.get(pk=self.kwargs.get("pk"))

    def get_queryset(self):
        return Member.all_objects.all()

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        responses = [
            r[1]
            for r in member_view.send_robust(self.get_object(), request=self.request)
        ]
        ctx["member_views"] = responses
        ctx["member"] = self.get_member()
        return ctx


class MemberListMixin:
    def get_members_queryset(self, search=None, _filter="active"):
        qs = Member.objects.all()
        if search:
            qs = qs.filter(Member.get_query_for_search(search))

        # Logic:
        #  + Active members have membership with start <= today and (end is null or end >= today)
        active_q = Q(memberships__start__lte=now().date()) & (
            Q(memberships__end__isnull=True) | Q(memberships__end__gte=now().date())
        )
        inactive_q = ~active_q
        if _filter == "all":
            pass
        elif _filter == "negbalance":
            return [m for m in qs.order_by("-id").distinct() if m.balance < 0]
        elif _filter == "inactive":
            qs = qs.filter(inactive_q)
        else:  # Default to 'active'
            qs = qs.filter(active_q)
        return qs.order_by("-id").distinct()


class MemberListView(MemberListMixin, ListView):
    template_name = "office/member/list.html"
    context_object_name = "members"
    model = Member
    paginate_by = 50

    def get_queryset(self):
        search = self.request.GET.get("q")
        _filter = self.request.GET.get("filter", "active")
        return self.get_members_queryset(search, _filter)

    def post(self, request, *args, **kwargs):
        for member in Member.objects.all():
            member.update_liabilites()
        return redirect(request.path)


class MemberDisclosureView(MemberListMixin, TemplateView):
    template_name = "office/member/disclosure.html"
    context_object_name = "members"
    model = Member

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["count"] = self.get_members_queryset().count()
        return context

    def post(self, request, *args, **kwargs):
        members = self.get_members_queryset()
        for member in members:
            member.record_disclosure_email.save()
        messages.success(
            request,
            _(
                "Generated {count} data disclosure emails and placed them in the outbox."
            ).format(count=len(members)),
        )
        return redirect("/members/list")


class MemberBalanceForm(forms.Form):
    start = forms.DateField(label=_("Start of balance timeframe"))
    end = forms.DateField(label=_("End of balance timeframe"))
    create_if_zero = forms.BooleanField(
        label=_("Create balances even if there was no payment due in the time chosen?"),
        required=False,
        initial=False,
    )
    balance_cutoff = forms.DecimalField(
        required=False,
        label=_("Balance cutoff"),
        help_text=_(
            "Only members with a deficit greater than this will receive an email."
        ),
        min_value=0,
    )
    subject = forms.CharField(label=_("Subject"))
    text = forms.CharField(
        label=_("Text"),
        initial=_(
            """Hello {name},

we wanted to let you know that within the time from {start}
to {end} you incurred unpaid liabilities in the amount of

   EUR {amount}

Please settle this amount as soon as possible, or let us
know if you think this email is incorrect."""
        ),
        widget=forms.Textarea,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["start"].widget.attrs["class"] = " datepicker"
        self.fields["end"].widget.attrs["class"] = " datepicker"
        # TODO: set start/end

    def clean(self):
        cleaned_data = super().clean()
        end = cleaned_data.get("end")
        start = cleaned_data.get("start")
        if end >= datetime.now().date():
            raise forms.ValidationError(_("End must be in the past!"))
        if end < start:
            raise forms.ValidationError(_("Start must be before end!"))
        return cleaned_data


class MemberBalanceView(MemberListMixin, FormView):
    template_name = "office/member/balance.html"
    context_object_name = "members"
    model = Member
    form_class = MemberBalanceForm

    def form_valid(self, form):
        members = Member.objects.with_active_membership()
        mails = errors = balance_count = 0
        start = datetime.combine(form.cleaned_data.get("start"), time(0, 0))
        end = datetime.combine(form.cleaned_data.get("end"), time(23, 59))
        create_if_zero = form.cleaned_data.get("create_if_zero")
        balance_cutoff = form.cleaned_data.get("balance_cutoff")
        text = form.cleaned_data.get("text")
        subject = form.cleaned_data.get("subject")
        for member in members:
            balance = None
            try:
                balance = member.create_balance(
                    start=start, end=end, create_if_zero=create_if_zero
                )
                balance_count += 1 if balance else 0
            except Exception:
                errors += 1
            if balance and (not balance_cutoff or balance.amount < -balance_cutoff):
                mail = EMail.objects.create(
                    to=member.email,
                    balance=balance,
                    text=text.format(
                        name=member.name, start=start, end=end, amount=balance.amount
                    ),
                    subject=subject,
                )
                mail.members.add(member)
                mails += 1
        message = str(
            _(
                "{balance_count} balances were created and {mail_count} reminder emails were placed in the outbox."
            ).format(balance_count=balance_count, mail_count=mails)
        )
        if errors:
            message += " " + str(
                _(
                    "{errors} balances could not be created due to errors. Presumably, these members already have balances overlapping this timeframe.".format(
                        errors=errors
                    )
                )
            )
        messages.success(self.request, message)
        return redirect("/members/list")


class MemberListExportForm(forms.Form):
    field_list = forms.MultipleChoiceField(
        choices=[], widget=forms.CheckboxSelectMultiple
    )
    member_filter = forms.ChoiceField(
        choices=[
            ("active", _("Active members")),
            ("negbalance", _("Members with negative balance")),
            ("inactive", _("Only inactive members")),
            ("all", _("All members")),
        ]
    )
    export_format = forms.ChoiceField(
        choices=[
            ("csv", _("CSV (Comma Separated Values)")),
            ("csv_de", _("CSV (Semicolon Separated Values, German Windows versions)")),
            # ('xlsx', _("XLSX (Excel)")),
        ]
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        fields = Member.get_fields()
        self.fields["field_list"].choices = [
            (f.field_id, f.name) for f in fields.values()
        ]
        self.fields["field_list"].initial = [
            f.field_id
            for f in fields.values()
            if f.registration_form.get("position", -1) > -1
        ]


class csv_excel_de(csv.excel):
    delimiter = ";"


def filter_excel_de(data):
    if isinstance(data, (float, Decimal)):
        return (
            f"{data:18,.2f}".replace(",", "_")
            .replace(".", ",")
            .replace("_", ".")
            .strip()
        )
    else:
        return data


class MemberListExportView(
    FormView, MemberListMixin, MultipleObjectMixin, MultipleObjectTemplateResponseMixin
):
    template_name = "office/member/list_export.html"
    context_object_name = "members"
    model = Member
    form_class = MemberListExportForm

    def dispatch(self, *args, **kwargs):
        self.object_list = self.get_queryset()
        return super().dispatch(*args, **kwargs)

    @transaction.atomic
    def form_valid(self, form):
        fields = Member.get_fields()
        selected_fields = form.cleaned_data["field_list"]
        header = collections.OrderedDict(
            [
                (f.field_id, f.name)
                for f in fields.values()
                if f.field_id in selected_fields
            ]
        )
        data = self.get_data(
            form,
            [
                (f.field_id, f.getter)
                for f in fields.values()
                if f.field_id in selected_fields
            ],
        )

        LogEntry.objects.create(
            content_type=None,
            object_id=0,
            user=self.request.user,
            action_type="byro.members.export",
            data={
                "filter": form.cleaned_data["member_filter"],
                "format": form.cleaned_data["export_format"],
                "fields": collections.OrderedDict(
                    [(f_id, str(f_name)) for (f_id, f_name) in header.items()]
                ),
            },
        )

        if form.cleaned_data["export_format"].startswith("csv"):
            return self.export_csv(
                header, data, csv_format=form.cleaned_data["export_format"]
            )

        return redirect(self.request.get_full_path())

    def export_csv(self, header, data, csv_format="default"):
        class EchoBOM:
            """Dummy, based on the Django docs.

            This one adds one feature: It outputs a Unicode BOM (Byte-
            Order Mark) as the first character.
            """

            def write(self, value):
                if not hasattr(self, "have_bom"):
                    self.have_bom = True
                    return "\ufeff" + value
                else:
                    return value

        def row_converter_de(row_):
            return {k: filter_excel_de(v) for (k, v) in row_.items()}

        pseudo_buffer = EchoBOM()
        writer = csv.DictWriter(
            pseudo_buffer,
            header.keys(),
            dialect={"csv_de": csv_excel_de}.get(csv_format, "excel"),
        )
        converter = {"csv_de": row_converter_de}.get(csv_format, lambda x: x)
        response = StreamingHttpResponse(
            (writer.writerow(converter(row)) for row in chain([header], data)),
            content_type="text/csv; charset=utf-8",
            charset="utf-8",
        )
        response[
            "Content-Disposition"
        ] = f'attachment; filename="members_{now().date()}.csv"'
        return response

    def get_data(self, form, field_mapping):
        qs = self.get_members_queryset(_filter=form.cleaned_data["member_filter"])
        for m in qs.all():
            yield {f_id: f_getter(m) for (f_id, f_getter) in field_mapping}


def get_member_list_importers():
    return [r[1] for r in member_list_importers.send_robust(sender=None)]


class MemberListImportForm(forms.Form):
    importer = forms.ChoiceField(
        choices=lambda: tuple(
            (i["id"], i["label"]) for i in get_member_list_importers()
        )
    )
    upload_file = forms.FileField()


class MemberListImportView(FormView):
    template_name = "office/member/list_import.html"
    context_object_name = "members"
    model = Member
    form_class = MemberListImportForm

    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)

    def form_invalid(self, form):
        print(form.errors.as_json())
        return super().form_invalid(form)

    @transaction.atomic
    def form_valid(self, form):
        importers = {i["id"]: i for i in get_member_list_importers()}

        importer = importers[form.cleaned_data["importer"]]
        sha256sum = hashlib.sha256()
        for chunk in form.cleaned_data["upload_file"].chunks():
            sha256sum.update(chunk)

        LogEntry.objects.create(
            content_type=None,
            object_id=0,
            user=self.request.user,
            action_type="byro.members.import",
            data={
                "importer": form.cleaned_data["importer"],
                "sha256": sha256sum.hexdigest(),
            },
        )

        return importer["form_valid"](self, form)


def get_encoding(form):
    detector = UniversalDetector()
    for chunk in form.cleaned_data["upload_file"].chunks():
        detector.feed(chunk)
        if detector.done:
            break
    detector.close()
    return detector.result["encoding"]


def create_membership(membership_parms, member):
    if membership_parms:
        for k in "start", "end":
            #  FIXME Also, we should find a way to handle dates in a generic way
            if membership_parms.get(k, None):
                membership_parms[k] = dateparser.parse(
                    membership_parms[k], languages=[settings.LANGUAGE_CODE, "en"]
                )
        Membership.objects.create(member=member, **membership_parms)


@transaction.atomic  # noqa
def default_csv_form_valid(view, form, dialect="excel"):
    mapping = None
    fields = Member.get_fields()
    encoding = get_encoding(form)

    with form.cleaned_data["upload_file"].open() as fp:
        instream = unicodecsv.DictReader(fp, dialect=dialect, encoding=encoding)

        for indict in instream:
            if mapping is None:
                mapping = {}
                for k in indict.keys():
                    if not k.strip():
                        continue
                    for field in fields.values():
                        if str(field.name).strip() == k.strip():
                            mapping[k.strip()] = field
                            break
                    else:
                        messages.error(
                            view.request,
                            _("Couldn't map input column '{}' to field").format(
                                k.strip()
                            ),
                        )
                        return redirect(view.request.get_full_path())

            do_update = False
            have_changes = False
            create_initial_balance = False
            initial_balance = None
            initial_balance_timestamp = None
            for verb_name, field in mapping.items():
                if field.field_id == "_internal_id":
                    member = Member.all_objects.filter(
                        pk=indict[verb_name.strip()]
                    ).first()
                    if member:
                        do_update = True
                        del indict[verb_name]

            if not do_update:
                member = Member.objects.create()

            membership_parms = {}
            for k, v in indict.items():
                if k.strip() not in mapping:
                    continue
                field = mapping[k.strip()]
                #  FIXME We're special casing the Membership here, but really Field.setter should handle that
                #  (In the case of 'member.memberships.last()' for the getter it should create a new Membership
                #  if no Membership exists.)
                if field.field_id.startswith("membership__"):
                    if v:
                        membership_parms[field.field_id.split("__", 1)[1]] = v
                else:
                    if do_update:
                        old_v = field.getter(member)
                        if old_v != v and not (old_v is None and not v):
                            field.setter(member, v)
                            have_changes = True
                    else:
                        if field.field_id == "_internal_balance":
                            if dialect == csv_excel_de:
                                v = v.replace(".", "").replace(",", ".")
                            initial_balance = Decimal(v)
                            if initial_balance != 0:
                                create_initial_balance = True
                        elif field.field_id == "_internal_last_transaction":
                            if v != "" and v is not None:
                                initial_balance_timestamp = dateparser.parse(
                                    v, languages=[settings.LANGUAGE_CODE, "en"]
                                )
                                create_initial_balance = True
                        else:
                            field.setter(member, v)

            if do_update:
                if have_changes:
                    member.log(view, ".updated")
                    member.save()
            else:
                if create_initial_balance and not initial_balance_timestamp:
                    messages.error(
                        view.request,
                        _(
                            "Either both or none columns has to be given: '{}' and '{}'"
                        ).format(
                            fields["_internal_balance"].name,
                            fields["_internal_last_transaction"].name,
                        ),
                    )
                    return redirect(view.request.get_full_path())

                member.log(view, ".created")
                member.save()

                # FIXME Changing membership when do_update is not implemented
                if membership_parms:
                    create_membership(membership_parms, member)

                if create_initial_balance:
                    balance_changed = member.adjust_balance(
                        view,
                        "Initial Balance created by CSV Import",
                        initial_balance,
                        SpecialAccounts.fees_receivable,
                        SpecialAccounts.opening_balance,
                        initial_balance_timestamp,
                    )

                    if balance_changed:
                        member.log(
                            view, ".finance.initial_balance", balance=member.balance
                        )

    return redirect(reverse("office:members.list"))


@receiver(member_list_importers)
def default_csv_importer(sender, **kwargs):
    return {
        "id": "byro.office.members.import.default_csv",
        "label": _("Generic CSV import (Comma Separated Values)"),
        "form_valid": default_csv_form_valid,
    }


@receiver(member_list_importers)
def default_csv_importer_german(sender, **kwargs):
    return {
        "id": "byro.office.members.import.default_csv_german",
        "label": _(
            "Generic CSV import (Semicolon Separated Values, German Windows versions)"
        ),
        "form_valid": partial(default_csv_form_valid, dialect=csv_excel_de),
    }


class MemberCreateView(FormView):
    template_name = "office/member/add.html"
    form_class = CreateMemberForm

    def get_object(self):
        return Member.objects.get(pk=self.kwargs["pk"])

    @transaction.atomic
    def form_valid(self, form):
        self.form = form
        form.save()
        messages.success(
            self.request,
            _("The member was added, please edit additional details if applicable."),
        )
        form.instance.log(self, ".created")

        responses = new_member.send_robust(sender=form.instance)
        for _receiver, response in responses:
            if isinstance(response, Exception):
                messages.warning(
                    self.request,
                    _("Some post processing steps could not be completed: ")
                    + str(response),
                )
        config = Configuration.get_solo()

        if config.welcome_member_template and form.instance.email:
            context = {
                "name": config.name,
                "contact": config.mail_from,
                "number": form.instance.number,
                "member_name": form.instance.name,
            }
            responses = [
                r[1]
                for r in new_member_mail_information.send_robust(sender=form.instance)
                if r
            ]
            context["additional_information"] = "\n".join(responses).strip()
            config.welcome_member_template.to_mail(
                email=form.instance.email, context=context
            )
        if config.welcome_office_template:
            context = {"member_name": form.instance.name}
            responses = [
                r[1]
                for r in new_member_office_mail_information.send_robust(
                    sender=form.instance
                )
                if r
            ]
            context["additional_information"] = "\n".join(responses).strip()
            config.welcome_office_template.to_mail(
                email=config.backoffice_mail, context=context
            )
        return super().form_valid(form)

    def get_context_data(self, **kwargs):
        result = super().get_context_data(**kwargs)
        config = Configuration.get_solo()
        result["name_config"] = {
            "order": config.default_order_name,
            "direct": config.default_direct_address_name,
        }
        return result

    def get_success_url(self):
        return reverse("office:members.data", kwargs={"pk": self.form.instance.pk})


class MemberDashboardView(MemberView):
    template_name = "office/member/dashboard.html"

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        obj = self.get_object()
        if not obj.memberships.count():
            return context
        first = obj.memberships.first().start
        delta = now().date() - first
        context["member_since"] = {
            "days": int(delta.total_seconds() / (60 * 60 * 24)),
            "years": int(round(delta.days / 365, 1)),
            "first": first,
        }
        context["current_membership"] = {
            "amount": obj.memberships.last().amount,
            "interval": obj.memberships.last().get_interval_display(),
        }
        context["statute_barred_debt"] = {"now": obj.statute_barred_debt()}
        context["statute_barred_debt"]["in1year"] = (
            obj.statute_barred_debt(relativedelta(years=1))
            - context["statute_barred_debt"]["now"]
        )

        context["tiles"] = []
        for __, response in member_dashboard_tile.send(self.request, member=obj):
            if not response:
                continue
            if isinstance(response, collections.abc.Mapping):
                context["tiles"].append(response)

        return context


class MemberDataView(MemberView):
    template_name = "office/member/data.html"

    def _instantiate(
        self,
        form_class,
        member,
        profile_class=None,
        instance=None,
        prefix=None,
        empty=False,
    ):
        params = {
            "instance": (
                getattr(
                    member, profile_class._meta.get_field("member").related_query_name()
                )
                if profile_class
                else instance
            )
            if not empty
            else None,
            "prefix": prefix
            or (
                profile_class.__name__
                if profile_class
                else instance.__class__.__name__ + "_"
                if instance
                else "member_"
            ),
            "data": self.request.POST if self.request.method == "POST" else None,
        }
        return form_class(**params)

    def get_forms(self):
        obj = self.get_object()
        membership_create_form = forms.modelform_factory(
            Membership, fields=["start", "end", "interval", "amount"]
        )
        for key in membership_create_form.base_fields:
            membership_create_form.base_fields[key].required = False
        return (
            [
                self._instantiate(
                    forms.modelform_factory(Member, exclude=["membership_type"]),
                    member=obj,
                    instance=obj,
                )
            ]
            + [
                self._instantiate(
                    forms.modelform_factory(Membership, exclude=["member"]),
                    member=obj,
                    instance=m,
                    prefix=m.id,
                )
                for m in obj.memberships.all()
            ]
            + [
                self._instantiate(
                    membership_create_form,
                    member=obj,
                    profile_class=Membership,
                    empty=True,
                )
            ]
            + [
                self._instantiate(
                    forms.modelform_factory(
                        profile_class,
                        fields=[
                            f.name
                            for f in profile_class._meta.fields
                            if f.name not in ["id", "member"]
                        ],
                    ),
                    member=obj,
                    profile_class=profile_class,
                )
                for profile_class in obj.profile_classes
            ]
        )

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["forms"] = self.get_forms()
        return context

    @transaction.atomic
    def post(self, *args, **kwargs):
        any_changed = False
        form_list = self.get_forms()
        for form in form_list:
            if not form.is_valid():
                return self.get(self.request)
        for form in form_list:
            if form.has_changed():
                if not getattr(form.instance, "member", False):
                    form.instance.member = self.get_object()
                any_changed = True
                form.save()
        if any_changed:
            self.get_object().log(self, ".updated")
            messages.success(self.request, _("Your changes have been saved."))
            update_member.send_robust(sender=self.request, form_list=form_list)
        return redirect(reverse("office:members.data", kwargs=self.kwargs))


class MemberFinanceView(MemberView):
    template_name = "office/member/finance.html"
    paginate_by = 50

    def get_bookings(self):
        account_list = [SpecialAccounts.donations, SpecialAccounts.fees_receivable]
        member = self.get_member()
        bookings = (
            Booking.objects.with_transaction_data()
            .filter(
                Q(debit_account__in=account_list) | Q(credit_account__in=account_list),
                member=member,
                transaction__value_datetime__lte=now(),
            )
            .order_by(
                "-transaction__value_datetime",
                "-booking_datetime",
                "-transaction__booking_datetime",
            )
        )
        if not member.balances.exists():
            return bookings
        result = []
        balances = list(member.balances.order_by("-start"))
        current_balance = balances.pop(0)
        current_balance.bookings = []
        for booking in bookings:
            while (
                current_balance
                and booking.transaction.value_datetime < current_balance.start
            ):
                result.append(current_balance)
                current_balance = balances.pop(0) if balances else None
                if current_balance:
                    current_balance.bookings = []
            if not current_balance:
                result.append(booking)
            elif booking.transaction.value_datetime > current_balance.end:
                result.append(booking)
            else:
                current_balance.bookings.append(booking)
        return result

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["member"] = self.get_member()
        context["bookings"] = self.get_bookings()
        return context


class MemberDocumentsView(MemberView, FormView):
    template_name = "office/member/documents.html"
    paginate_by = 50
    form_class = DocumentUploadForm

    def post(self, *args, **kwargs):
        self.object = self.get_object()
        return super().post(*args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["member"] = self.get_member()
        return context

    def get_form(self):
        form = super().get_form()
        return form

    def get_success_url(self):
        return reverse("office:members.documents", kwargs={"pk": self.get_member().pk})

    @transaction.atomic
    def form_valid(self, form):
        self.form = form
        member = self.get_member()

        form.instance.member = member
        form.save()
        member.log(
            self,
            ".document.created",
            document=form.instance,
            content_hash=form.instance.content_hash,
        )

        return super().form_valid(form)


class MemberAccountAdjustmentForm(forms.Form):
    form_title = _("Adjust member account balance")

    date = forms.DateField(initial=lambda: now().date())
    adjustment_reason = forms.ChoiceField(
        choices=[("initial", _("Initial balance")), ("waiver", _("Fees waived"))]
    )
    adjustment_memo = forms.CharField(required=False)
    adjustment_type = forms.ChoiceField(
        widget=forms.RadioSelect,
        choices=[
            ("relative", _("Relative (Add or subtract amount to/from balance)")),
            ("absolute", _("Absolute (Balance should be)")),
        ],
        initial="relative",
    )
    amount = forms.DecimalField(initial=Decimal("0.00"), decimal_places=2)

    date.widget.attrs.update({"class": "datepicker"})


class MultipleFormsMixin:
    def get_operations(self):
        raise NotImplementedError

    def mangle_button(self, name, prefix):
        return f"submit_{prefix}_{name}"

    def get_forms(self):
        """Instantiate forms, return a list of tuples like get_operations(),
        but with Form objects and expanded prefix values."""
        retval = []

        for prefix, title, form_class, buttons, callback in self.get_operations():
            retval.append(
                (
                    prefix,
                    title,
                    form_class(
                        prefix=prefix,
                        data=self.request.POST
                        if self.request.method == "POST"
                        else None,
                    ),
                    buttons,
                    callback,
                )
            )

        return retval

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["forms"] = self.get_forms()
        return context

    def post(self, *args, **kwargs):
        retval = None

        for prefix, _title, form, buttons, callback in self.get_forms():
            active_buttons = [
                name
                for name in buttons
                if self.mangle_button(name, prefix) in self.request.POST
            ]
            if active_buttons:
                if form.is_valid():
                    retval = callback(form, active_buttons) or retval

        if retval:
            return retval

        return redirect(self.request.get_full_path())


class MemberOperationsView(MultipleFormsMixin, MemberView):
    template_name = "office/member/operations.html"
    membership_form_class = forms.modelform_factory(
        Membership, fields=["start", "end", "interval", "amount"]
    )

    def get_operations(self):
        """Return a list of tuples. Each one:

        + internal name/prefix of the form + Title of the form + A
        callable returning a Form instance + (Ordered)Dict of submit
        buttons {button_name: button_text} + Callback function for
        successful submit of the form
        """
        member = self.get_object()
        now_ = now()

        retval = []

        # Add Leave forms for all current memberships
        def _create_ms_leave_form(*args, **kwargs):
            f = self.membership_form_class(instance=ms, *args, **kwargs)
            f.fields["start"].disabled = True
            f.fields["interval"].disabled = True
            f.fields["amount"].disabled = True
            f.fields["end"].required = True
            f.fields["end"].widget.attrs["class"] = "datepicker"
            return f

        for ms in member.memberships.all().order_by("-start"):
            if ms.start <= now_.date() and (not ms.end or ms.end > now_.date()):
                retval.append(
                    (
                        f"ms_{ms.pk}_leave",
                        _("End membership"),
                        _create_ms_leave_form,
                        {"end": _("End membership")},
                        lambda *args, **kwargs: self.end_membership(
                            ms, *args, **kwargs  # noqa
                        ),
                    )
                )

        # Add account adjustment form
        retval.append(
            (
                "member_account_adjustment",
                _("Adjust member account balance"),
                MemberAccountAdjustmentForm,
                {"adjust": _("Adjust balance")},
                self.adjust_balance,
            )
        )

        return retval

    @transaction.atomic
    def adjust_balance(self, form, active_buttons):
        memo = (
            form.cleaned_data["adjustment_memo"]
            or dict(form.fields["adjustment_reason"].choices).get(
                form.cleaned_data["adjustment_reason"], None
            )
            or _("Account adjustment")
        )

        member = self.get_member()

        if form.cleaned_data["adjustment_type"] == "relative":
            amount = form.cleaned_data["amount"]
        else:
            old_balance = member._calc_balance(
                form.cleaned_data["date"], form.cleaned_data["date"]
            )
            amount = old_balance - form.cleaned_data["amount"]

        amount_, from_, to_ = None, None, None

        if form.cleaned_data["adjustment_reason"] == "initial":
            amount_, from_, to_ = (
                amount,
                SpecialAccounts.opening_balance,
                SpecialAccounts.fees_receivable,
            )
        elif form.cleaned_data["adjustment_reason"] == "waiver":
            if amount < 0:
                amount_, from_, to_ = (
                    -amount,
                    SpecialAccounts.fees_receivable,
                    SpecialAccounts.lost_income,
                )
            elif amount > 0:
                messages.error(
                    self.request,
                    _(
                        "Fee waiving needs to decrease debts. Use a negative value in relative mode, or a value higher than the current one in absolute mode."
                    ),
                )
                return

        balance_changed = member.adjust_balance(
            self, memo, amount_, from_, to_, form.cleaned_data["date"]
        )

        if balance_changed:
            balance = member.balance

            if form.cleaned_data["adjustment_reason"] == "initial":
                member.log(self, ".finance.initial_balance", balance=balance)
            elif form.cleaned_data["adjustment_reason"] == "waiver":
                member.log(self, ".finance.fees_waived", amount=amount)
            else:
                member.log(
                    self, ".finance.account_adjusted", balance=balance, amount=amount
                )

            messages.success(
                self.request,
                _(
                    "Membership account adjusted by {amount}, current balance is {balance}"
                ).format(amount=amount, balance=balance),
            )

    @transaction.atomic
    def end_membership(self, ms, form, active_buttons):
        if form.instance.end:
            if not getattr(form.instance, "member", False):
                form.instance.member = self.get_object()

            form.save()
            form.instance.log(self, ".ended")
            messages.success(
                self.request,
                _(
                    "The membership has been terminated. Please check the outbox for the notifications."
                ),
            )

            form.instance.member.update_liabilites()

            responses = leave_member.send_robust(sender=form.instance)
            for _receiver, response in responses:
                if isinstance(response, Exception):
                    messages.warning(
                        self.request,
                        _("Some post processing steps could not be completed: ")
                        + str(response),
                    )

            config = Configuration.get_solo()
            if config.leave_member_template:
                context = {
                    "name": config.name,
                    "contact": config.mail_from,
                    "number": form.instance.member.number,
                    "member_name": form.instance.member.name,
                    "end": form.instance.end,
                }
                responses = [
                    r[1]
                    for r in leave_member_mail_information.send_robust(
                        sender=form.instance
                    )
                    if r
                ]
                context["additional_information"] = "\n".join(responses).strip()
                config.leave_member_template.to_mail(
                    email=form.instance.member.email, context=context
                )
            if config.leave_office_template:
                context = {
                    "member_name": form.instance.member.name,
                    "end": form.instance.end,
                }
                responses = [
                    r[1]
                    for r in leave_member_office_mail_information.send_robust(
                        sender=form.instance
                    )
                    if r
                ]
                context["additional_information"] = "\n".join(responses).strip()
                config.leave_office_template.to_mail(
                    email=config.backoffice_mail, context=context
                )


class MemberListTypeaheadView(View):
    def dispatch(self, request, *args, **kwargs):
        search = request.GET.get("search")
        if not search or len(search) < 2:
            return JsonResponse({"count": 0, "results": []})

        queryset = Member.objects.filter(Member.get_query_for_search(search))
        return JsonResponse(
            {
                "count": len(queryset),
                "results": [
                    {
                        "id": member.pk,
                        "nick": member.profile_profile.nick,
                        "name": member.name,
                    }
                    for member in queryset
                ],
            }
        )


class MemberRecordDisclosureView(MemberView):
    template_name = "office/member/data_disclosure.html"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["mail"] = self.get_member().record_disclosure_email
        return ctx

    def post(self, request, *args, **kwargs):
        self.get_member().record_disclosure_email.save()
        self.get_member().log(self, ".disclosure_email_generated")
        messages.success(
            request, _("The email was generated and can be sent in the outbox.")
        )
        return redirect(reverse("office:members.dashboard", kwargs=self.kwargs))


class MemberLogView(MemberView):
    template_name = "office/member/log.html"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["log_entries"] = self.get_member().log_entries()
        return ctx


class MemberMailsView(MemberView):
    template_name = "office/member/mails.html"


class MemberTimelineView(MemberView):
    template_name = "office/member/timeline.html"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["timeline"] = self.get_timeline()
        return ctx

    def get_timeline(self):
        member = self.get_member()

        entries = sorted_merge(
            get_finance_timeline(member),
            get_mail_timeline(member),
            get_ops_timeline(member),
            get_document_timeline(member),
        )

        return augment_timeline(entries)



# Generated by Django 2.1.7 on 2019-02-26 21:38

import byro.public.models
import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("public", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="memberpageprofile",
            name="publication_consent",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True, default=byro.public.models.get_default_consent, null=True
            ),
        )
    ]



import base64
import datetime
import decimal
import uuid
from functools import wraps

import canonicaljson
import nacl.hash
import nacl.utils
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import F, Q
from django.db.models.signals import pre_delete, pre_save
from django.dispatch import receiver
from django.utils.timezone import now

from byro.common.utils import get_installed_software


class ContentObjectManager(models.Manager):
    """An object manager that can handle filter(content_object=...)"""

    def filter(self, *args, **kwargs):
        if "content_object" in kwargs:
            content_object = kwargs.pop("content_object")
            kwargs["content_type"] = ContentType.objects.get_for_model(
                type(content_object)
            )
            kwargs["object_id"] = content_object.pk
        return super().filter(*args, **kwargs)


class LogEntryManager(ContentObjectManager):
    """Manager that is linking the log chain on .create()"""

    def create(self, *args, **kwargs):
        kwargs["auth_prev"] = self.get_chain_end()
        return super().create(*args, **kwargs)

    def get_chain_end(self):
        return self.filter(Q(auth_next=None) | Q(auth_prev_id=F("auth_hash"))).first()


PERSON_BYTES = b"byro logchain v1"


class LogEntry(models.Model):
    """A log entry for a change (or addition, or modification) in the database.

    :param content_object: The object that was modified (or added)
    :param datetime: Timestamp of the logged action
    :param user: The user that performed the action
    :param action_type: dot-namespaced type of the action, such as "byro.office.members.added"
    :param data: arbitrary additional data about the action.
      Well-known keys:
        * source: Actor/initiator of the change (a user or an automatic process). This must always be present. Will be generated from user attribute if set.
        * old: Value before the change
        * new: Value after the change
        * changes: dictionary of changes  key: (old_value, new_value)
    :param auth_hash: Authentication hash for the log chain. See docs/administrator/log-chain.rst
    :param auth_prev: `auth_hash` of the previous entry.
    :param auth_data: Authentication data for the log chain:
        * hash_ver: hash/algorithm version
        * nonce: A random nonce
        * data_mac: Message Authentication Code for the LogEntry.data attribute.
        * orig_content_type: Copy of the ContentType value at LogEntry creation
        * orig_user_id: Copy of the user_id value at LogEntry creation
    """

    content_type = models.ForeignKey(ContentType, null=True, on_delete=models.SET_NULL)
    object_id = models.PositiveIntegerField(db_index=True)
    content_object = GenericForeignKey("content_type", "object_id")
    datetime = models.DateTimeField(default=now, db_index=True)
    user = models.ForeignKey(get_user_model(), null=True, on_delete=models.SET_NULL)

    action_type = models.CharField(max_length=255)
    data = models.JSONField(null=True)

    auth_hash = models.CharField(max_length=140, null=False, unique=True)
    auth_prev = models.ForeignKey(
        "self",
        on_delete=models.PROTECT,
        related_name="auth_next",
        to_field="auth_hash",
        null=False,
        blank=False,
    )
    auth_data = models.JSONField(null=False)

    objects = LogEntryManager()

    class Meta:
        ordering = ("-datetime", "-id")
        indexes = [
            models.Index(fields=["content_type", "object_id"]),
            models.Index(fields=["action_type"]),
        ]

    def delete(self, using=None, keep_parents=False):
        raise TypeError("Logs cannot be deleted.")

    def save(self, *args, **kwargs):
        if kwargs.get("update_fields", None) == [
            "auth_hash"
        ] and self.auth_hash.startswith("random:"):
            self._OVERRIDE_SAVE = ["auth_hash"]
        elif (
            kwargs.get("update_fields", None) == ["auth_prev"]
            and self.auth_prev == "undefined:0"
        ):
            self._OVERRIDE_SAVE = ["auth_prev"]
        else:
            if getattr(self, "pk", None):
                raise TypeError("Logs cannot be modified.")

            if not self.data:
                self.data = {}

            if self.user:
                self.data.setdefault("source", str(self.user))

            if not self.data.get("source"):
                raise ValueError("Need to provide at least user or data['source']")

            self._compute_logchain()

        return super().save(*args, **kwargs)

    def _compute_logchain(self):
        hashed_data_dict = dict(self.data)
        hdd_encoded = canonicaljson.encode_canonical_json(hashed_data_dict)

        hdd_nonce = nacl.utils.random(12)
        hdd_mac = nacl.hash.blake2b(hdd_encoded, digest_size=64, salt=hdd_nonce)

        if not self.datetime:
            self.datetime = now()

        self.auth_data = {
            "hash_ver": 2,
            "nonce": base64.b64encode(hdd_nonce).decode("us-ascii"),
            "data_mac": "blake2b:{}".format(hdd_mac.decode("us-ascii")),
            "orig_content_type": "{}.{}".format(
                self.content_type.app_label, self.content_type.model
            )
            if self.content_type
            else None,
            "orig_user_id": self.user_id if self.user_id else None,
            "software_version": ", ".join(get_installed_software()),
        }

        authenticated_dict = self.get_authenticated_dict()

        ad_encoded = canonicaljson.encode_canonical_json(authenticated_dict)
        self.auth_hash = "blake2b:{}".format(
            nacl.hash.blake2b(ad_encoded, digest_size=64, person=PERSON_BYTES).decode(
                "us-ascii"
            )
        )

        if not self.auth_prev_id:
            self.auth_prev_id = self.auth_hash

    def get_authenticated_dict(self):
        return {
            "object_id": str(self.object_id)
            if not isinstance(self.object_id, (int, str))
            else self.object_id,
            "datetime": self.datetime.isoformat(),
            "action_type": self.action_type,
            "prev_hash": self.auth_prev_id
            if self.auth_prev_id and self.auth_prev_id != self.auth_hash
            else "initial:0",
            "auth_data": self.auth_data,
            "source": self.data["source"],
        }

    def verify(self):
        if self.auth_data["hash_ver"] not in (1, 2):
            return False

        if (
            self.content_type is not None
            and f"{self.content_type.app_label}.{self.content_type.model}"
            != self.auth_data["orig_content_type"]
        ):
            return False

        if self.user_id is not None and self.user_id != self.auth_data["orig_user_id"]:
            return False

        hashed_data_dict = dict(self.data)
        hdd_encoded = canonicaljson.encode_canonical_json(hashed_data_dict)

        hdd_nonce = base64.b64decode(self.auth_data["nonce"])
        hdd_mac = nacl.hash.blake2b(hdd_encoded, digest_size=64, salt=hdd_nonce)

        if (
            "blake2b:{}".format(hdd_mac.decode("us-ascii"))
            != self.auth_data["data_mac"]
        ):  # FIXME Maybe fixed time comparison?
            return False

        authenticated_dict = self.get_authenticated_dict()

        ad_encoded = canonicaljson.encode_canonical_json(authenticated_dict)
        auth_hash = "blake2b:{}".format(
            nacl.hash.blake2b(ad_encoded, digest_size=64, person=PERSON_BYTES).decode(
                "us-ascii"
            )
        )

        return auth_hash == self.auth_hash  # FIXME Maybe fixed time comparison?


@receiver(pre_save, sender=LogEntry)
def log_entry_pre_save(sender, instance, *args, **kwargs):
    if getattr(instance, "_OVERRIDE_SAVE", None) == [
        "auth_hash"
    ] and instance.auth_hash.startswith("random:"):
        delattr(instance, "_OVERRIDE_SAVE")
    elif (
        getattr(instance, "_OVERRIDE_SAVE", None) == ["auth_prev"]
        and instance.auth_prev == "undefined:0"
    ):
        delattr(instance, "_OVERRIDE_SAVE")
    elif instance.pk:
        raise TypeError("Logs cannot be modified.")


@receiver(pre_delete, sender=LogEntry)
def log_entry_pre_delete(sender, instance, *args, **kwargs):
    if instance.pk:
        raise TypeError("Logs cannot be deleted.")


def flatten_objects(inobj, key_was=None):
    if isinstance(inobj, dict):
        return {k: flatten_objects(v, key_was=k) for k, v in inobj.items()}
    elif isinstance(inobj, (tuple, list)):
        return [flatten_objects(v) for v in inobj]
    elif isinstance(inobj, datetime.datetime):
        return inobj.strftime("%Y-%m-%d %H:%M:%S %Z")
    elif isinstance(inobj, datetime.date):
        return inobj.strftime("%Y-%m-%d")
    elif isinstance(inobj, uuid.UUID):
        return str(inobj)
    elif isinstance(inobj, decimal.Decimal) or (
        key_was == "amount" and isinstance(inobj, (int, float))
    ):
        return f"{inobj:.2f}"
    else:
        try:
            content_type = ContentType.objects.get_for_model(type(inobj))
            return {
                "object": content_type.name,
                "ref": (content_type.app_label, content_type.model, inobj.pk),
                "value": str(inobj),
            }
        except Exception:
            return str(inobj)


class LogTargetMixin:
    LOG_TARGET_BASE = None

    def log(self, context, action, user=None, **kwargs):
        if hasattr(context, "request"):
            context = context.request

        user = user or getattr(context, "user", None)

        if isinstance(context, str) and "source" not in kwargs:
            kwargs["source"] = context

        if self.LOG_TARGET_BASE and action.startswith("."):
            action = self.LOG_TARGET_BASE + action

        kwargs = flatten_objects(kwargs)

        LogEntry.objects.create(
            content_object=self, user=user, action_type=action, data=dict(kwargs)
        )

    def log_entries(self):
        return LogEntry.objects.filter(content_object=self)


def log_call(action, log_on="retval"):
    def outer_decorator(f):
        @wraps(f)
        def decorator(*args, **kwargs):
            if "user_or_context" not in kwargs:
                raise TypeError(
                    "You need to provide a 'user_or_context' named parameter which indicates the responsible user (a User model object), request (a View instance or HttpRequest object), or generic context (a str)."
                )
            user_or_context = kwargs.pop("user_or_context")
            user = kwargs.pop("user", None)

            retval = f(*args, **kwargs)

            log_kwargs = {k: v for k, v in kwargs.items() if v}
            log_args = list(args)
            log_args = log_args[
                1:
            ]  # Warning: we assume that args[0] is 'self'. Only works correctly with calls to bound methods
            if log_args:
                log_kwargs["_args"] = [v for v in log_args]

            if log_on == "retval":
                retval.log(user_or_context, action, user=user, **log_kwargs)
            else:
                args[0].log(user_or_context, action, user=user, **log_kwargs)

            return retval

        return decorator

    return outer_decorator



# Generated by Django 1.11.6 on 2017-10-12 19:15

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("profile", "0001_initial")]

    operations = [
        migrations.RemoveField(model_name="memberprofile", name="address"),
        migrations.RemoveField(model_name="memberprofile", name="member_identifier"),
        migrations.RemoveField(model_name="memberprofile", name="name"),
        migrations.AddField(
            model_name="memberprofile",
            name="phone_number",
            field=models.CharField(blank=True, max_length=32, null=True),
        ),
    ]



import django.dispatch

new_member_mail_information = django.dispatch.Signal()
"""
Receives the new member as signal. Response will be added to the email
welcoming the new member.
"""
new_member_office_mail_information = django.dispatch.Signal()
"""
Receives the new member as signal. Response will be added to the email
notifying the office about the new member.
"""
new_member = django.dispatch.Signal()
"""
Receives the new member as signal. If an exception is raised, the error
message will be displayed in the frontend as a warning.
"""
leave_member_mail_information = django.dispatch.Signal()
"""
Receives the leaving member as signal. Response will be added to the email
confirming termination to the member.
"""
leave_member_office_mail_information = django.dispatch.Signal()
"""
Receives the leaving member as signal. Response will be added to the email
notifying the office about the termination of the member.
"""
leave_member = django.dispatch.Signal()
"""
Receives the new member as signal. If an exception is raised, the error
message will be displayed in the frontend as a warning.
"""
update_member = django.dispatch.Signal()
"""
If a member is updated via the office form collection at members/view/{id}/data.
The signal receives the request, and the form_list as parameters. The changes
will already have been saved at this point.
"""



import django.dispatch

member_view = django.dispatch.Signal()
"""
This signal allows you to add a tab to the member detail view tab list.
Receives the member as sender, and additionally the request
Must return a dict::

    {
        "label": _("Fancy Member View"),
        "url": "/member/123/foo/",
        "url_name": "plugins:myplugin:foo_view",
    }

Please use ``byro.office.views.members.MemberView`` as base class for these views.
"""

nav_event = django.dispatch.Signal()
"""
This signal allows you to add additional views to the sidebar.
Receives the request as sender. Must return a dictionary containing at least
the keys ``label`` and ``url``. You can also return a ForkAwesome icon name
with the key ``icon``. You should also return an ``active`` key with a boolean
set to ``True`` if this item should be marked as active.

If you want your Plugin to appear in the "Finance" or "Settings" submenu in the
side bar, please set ``section`` in your return dict to either ``finance`` or
``settings``, and don't set an ``icon``.

May return an iterable of multiple dictionaries as described above.
"""

member_list_importers = django.dispatch.Signal()
"""
This signal allows you to add additional member list importers.
Receives None as argument, must return a dict::

    {
        "id": "dot.scoped.importer.id",
        "label": _("My super importer"),
        "form_valid": form_valid_callback,
    }

where `form_valid_callback` should accept two arguments: view (the View object
handling the request), and form (the form object that was submitted, the file
to import is in the `upload_file` form field) and should return a Response
object.
"""
member_dashboard_tile = django.dispatch.Signal()
"""
This signal allows you to add tiles to the member's dashboard.
Receives None as argument, must return either None or a dict::

    {
        "title": _("Dash!"),
        "lines": [_('Line 1'), _('Line 2')]
        "url": "/member/123/foo/",
        "public": False,  # False is the default
    }

All of the parts of this dict are optional. You cannot include HTML in the
response, all strings will be escaped at render time. If "public" is set to
True, the dashboard tile will also be shown on the member's personal page.
"""



import pytest

from byro.members.models import Member


@pytest.fixture
def member_bob():
    member = Member.objects.create(email="bob@hacker.space")
    yield member
    [profile.delete() for profile in member.profiles]
    member.delete()


@pytest.mark.django_db
def test_sepa_models(member_bob):
    assert member_bob.profile_sepa.iban is None



# Generated by Django 1.11.6 on 2017-10-13 14:36

import annoying.fields
from django.db import migrations
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [("sepa", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="membersepa",
            name="member",
            field=annoying.fields.AutoOneToOneField(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="profile_sepa",
                to="members.Member",
            ),
        )
    ]



# Generated by Django 2.1.1 on 2018-09-29 09:15

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("mails", "0006_auto_20180911_1956")]

    operations = [
        migrations.AlterField(
            model_name="email",
            name="members",
            field=models.ManyToManyField(related_name="emails", to="members.Member"),
        )
    ]



# Generated by Django 2.1.1 on 2018-10-20 10:29

import django.contrib.postgres.fields.jsonb
import django.utils.timezone
from django.db import migrations, models

from uuid import uuid4


def fill_hash(apps, schema_editor):
    LogEntry = apps.get_model("common", "LogEntry")
    for entry in LogEntry.objects.filter(auth_hash__isnull=True).all():
        entry.auth_hash = f"random:{uuid4().hex}"
        entry.save(update_fields=["auth_hash"])


def fill_prev(apps, schema_editor):
    LogEntry = apps.get_model("common", "LogEntry")
    prev = None
    for entry in (
        LogEntry.objects.filter(auth_prev="undefined:0")
        .order_by("datetime", "id")
        .all()
    ):
        entry.auth_prev = prev or entry
        prev = entry
        entry.save(update_fields=["auth_prev"])


class Migration(migrations.Migration):
    dependencies = [("common", "0012_auto_20180929_1317")]

    operations = [
        # https://stackoverflow.com/questions/28429933/django-migrations-using-runpython-to-commit-changes/39541048#39541048
        # UNCOMMENT this section if you are running a large byro upgrade that fails on migration
        # ALSO uncomment the section at the end of the list of operations
        # As an alternative, consider upgrading in smaller steps.
        # migrations.RunSQL(
        #     "SET CONSTRAINTS ALL IMMEDIATE", reverse_sql=migrations.RunSQL.noop
        # ),
        migrations.AlterField(
            model_name="logentry",
            name="datetime",
            field=models.DateTimeField(
                db_index=True, default=django.utils.timezone.now
            ),
        ),
        migrations.AddField(
            model_name="logentry",
            name="auth_data",
            field=django.contrib.postgres.fields.jsonb.JSONField(default={}),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="logentry",
            name="auth_hash",
            field=models.CharField(null=True, max_length=70),
        ),
        migrations.RunPython(fill_hash, migrations.RunPython.noop),
        migrations.AlterField(
            model_name="logentry",
            name="auth_hash",
            field=models.CharField(max_length=140, unique=True, null=False),
        ),
        migrations.AddField(
            model_name="logentry",
            name="auth_prev",
            field=models.ForeignKey(
                default="undefined:0",
                on_delete=django.db.models.deletion.PROTECT,
                related_name="auth_next",
                to="common.LogEntry",
                to_field="auth_hash",
                null=False,
                blank=False,
            ),
            preserve_default=False,
        ),
        migrations.RunPython(fill_prev, migrations.RunPython.noop),
        migrations.AddIndex(
            model_name="logentry",
            index=models.Index(
                fields=["content_type", "object_id"],
                name="common_loge_content_43b532_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="logentry",
            index=models.Index(
                fields=["action_type"], name="common_loge_action__1810d9_idx"
            ),
        ),
        # migrations.RunSQL(
        #     migrations.RunSQL.noop, reverse_sql="SET CONSTRAINTS ALL IMMEDIATE"
        # ),
    ]



import string
from urllib.parse import urljoin

from annoying.fields import AutoOneToOneField
from django.conf import settings
from django.db import models
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _

from byro.common.models.configuration import Configuration


def generate_default_token():
    return get_random_string(
        allowed_chars=string.ascii_lowercase + string.digits, length=32
    )


def get_default_consent():
    return {"fields": dict()}


class MemberpageProfile(models.Model):
    form_title = _("Memberpage settings")

    member = AutoOneToOneField(
        to="members.Member", on_delete=models.CASCADE, related_name="profile_memberpage"
    )
    secret_token = models.CharField(
        max_length=128,
        null=True,
        blank=True,
        default=generate_default_token,
        unique=True,
    )
    is_visible_to_members = models.BooleanField(
        default=False, verbose_name=_("Consent: Visible to other members")
    )
    # publication_consent format: {"fields": {"profile_memberpage__secret_token": {"visibility": "share"}}}
    publication_consent = models.JSONField(
        default=get_default_consent, null=True, blank=True
    )

    def get_url(self):
        config = Configuration.get_solo()
        relative_url = reverse(
            "public:memberpage:member.dashboard",
            kwargs={"secret_token": self.secret_token},
        )
        if config.public_base_url:
            return urljoin(config.public_base_url, relative_url)
        else:
            return urljoin(settings.SITE_URL, relative_url)

    def get_public_data(self):
        result = []
        if not self.is_visible_to_members or not self.publication_consent:
            return result
        all_fields = self.member.get_fields()
        for key, value in self.publication_consent.get("fields", {}).items():
            if not value.get("visibility") == "share" or key not in all_fields:
                continue
            field = all_fields[key]
            result.append(
                {"label": field.base_name, "value": field.getter(self.member)}
            )
        return result



# Generated by Django 1.11.6 on 2017-10-13 11:20

import byro.common.models.auditable
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [("members", "0002_auto_20171012_1857")]

    operations = [
        migrations.CreateModel(
            name="Membership",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("start", models.DateField(verbose_name="start")),
                ("end", models.DateField(blank=True, null=True, verbose_name="end")),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2,
                        help_text="The amount to be paid in the chosen interval",
                        max_digits=8,
                        verbose_name="amount",
                    ),
                ),
                (
                    "interval",
                    models.IntegerField(
                        choices=[
                            (1, "monthly"),
                            (3, "quarterly"),
                            (6, "biannually"),
                            (12, "annually"),
                        ],
                        help_text="How often does the member pay their fees?",
                        verbose_name="interval",
                    ),
                ),
                (
                    "member",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="memberships",
                        to="members.Member",
                    ),
                ),
            ],
            bases=(byro.common.models.auditable.Auditable, models.Model),
        ),
        migrations.CreateModel(
            name="MembershipType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=200, verbose_name="name")),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2,
                        help_text="Please enter the yearly fee for this membership type.",
                        max_digits=8,
                        verbose_name="amount",
                    ),
                ),
            ],
            bases=(byro.common.models.auditable.Auditable, models.Model),
        ),
    ]



# Generated by Django 2.1.1 on 2018-09-29 14:48

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
        ("common", "0010_auto_20180929_1052"),
    ]

    operations = [
        migrations.CreateModel(
            name="LogEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveIntegerField(db_index=True)),
                ("datetime", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("action_type", models.CharField(max_length=255)),
                ("data", django.contrib.postgres.fields.jsonb.JSONField(null=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="contenttypes.ContentType",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"ordering": ("-datetime", "-id")},
        )
    ]



from django.contrib.auth import get_user_model
from django.db import models


class Auditable:
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    created_by = models.ForeignKey(
        to=get_user_model(),
        on_delete=models.PROTECT,
        related_name="+",  # no related lookup
        null=True,
    )
    modified_by = models.ForeignKey(
        to=get_user_model(),
        on_delete=models.PROTECT,
        related_name="+",  # no related lookup
        null=True,
    )



from django.urls import include, re_path

from . import views

member_pages = [
    re_path(r"^$", views.MemberView.as_view(), name="member.dashboard"),
    re_path(r"^list$", views.MemberListView.as_view(), name="member.list"),
]

urlpatterns = [
    re_path(r"^member/(?P<secret_token>[^/]+)/", include((member_pages, "memberpage")))
]
app_name = "public"



import os

from django import forms
from django.contrib import messages
from django.db import transaction
from django.http import FileResponse
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, FormView

from byro.documents.models import Document, get_document_category_names


class DocumentUploadForm(forms.ModelForm):
    class Meta:
        model = Document
        exclude = ("content_hash", "member")

    def __init__(self, *args, **kwargs):
        initial_category = kwargs.pop("initial_category", "byro.documents.misc")

        super().__init__(*args, **kwargs)

        categories = get_document_category_names()

        self.fields["category"] = forms.ChoiceField(
            choices=sorted(categories.items()), initial=initial_category
        )
        if "class" in self.fields["date"].widget.attrs:
            self.fields["date"].widget.attrs["class"] += " datepicker"
        else:
            self.fields["date"].widget.attrs["class"] = "datepicker"


class DocumentUploadView(FormView):
    template_name = "office/documents/add.html"
    model = Document
    form_class = DocumentUploadForm

    @transaction.atomic
    def form_valid(self, form):
        form.save()
        form.instance.log(
            self,
            ".saved",
            file_name=form.instance.document.name,
            content_size=form.instance.document.size,
            content_hash=form.instance.content_hash,
        )
        try:
            messages.success(self.request, _("The upload was processed successfully."))
        except Exception as e:
            messages.error(
                self.request,
                _("The upload was added successfully, but could not be processed: ")
                + str(e),
            )
        self.form = form
        return super().form_valid(form)

    def get_success_url(self):
        return self.request.path


class DocumentDetailView(DetailView):
    template_name = "office/documents/detail.html"
    model = Document
    context_object_name = "document"


class DocumentDownloadView(DetailView):
    model = Document
    context_object_name = "document"

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        # HACK HACK
        # FileResponse can work with a file stream, and do optimized sending with some WSGI servers
        # but the FieldFile in a FileField will be closed upon return from this function, while
        # the wsgi.file_wrapper will operate after returning from Django.
        # dup() the open file, then return an fdopen()
        # FIXME: Fallback to another mechanism for non-local media storage
        response_file = os.fdopen(os.dup(self.object.document.fileno()), mode="rb")
        response = FileResponse(
            response_file,
            content_type=self.object.mime_type_guessed,
            filename=self.object.basename,
        )
        # The actions above may have seeked the fd
        response_file.seek(0)
        return response



# Generated by Django 1.11.6 on 2017-10-13 14:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("members", "0003_membership_membershiptype")]

    operations = [
        migrations.AlterField(
            model_name="member",
            name="address",
            field=models.TextField(blank=True, max_length=300, null=True),
        )
    ]



# Generated by Django 1.11.11 on 2018-03-31 10:21

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("members", "0006_auto_20180113_1849")]

    operations = [
        migrations.AddField(
            model_name="member",
            name="membership_type",
            field=models.CharField(default="member", max_length=40),
        )
    ]



import atexit
import os
import tempfile
from contextlib import suppress

from byro.settings import *  # noqa

tmpdir = tempfile.TemporaryDirectory()
os.environ.setdefault("DATA_DIR", tmpdir.name)


BASE_DIR = tmpdir.name
DATA_DIR = tmpdir.name
LOG_DIR = os.path.join(DATA_DIR, "logs")
MEDIA_ROOT = os.path.join(DATA_DIR, "media")
STATIC_ROOT = os.path.join(DATA_DIR, "static")
HTMLEXPORT_ROOT = os.path.join(DATA_DIR, "htmlexport")

for directory in (BASE_DIR, DATA_DIR, LOG_DIR, MEDIA_ROOT, HTMLEXPORT_ROOT):
    os.makedirs(directory, exist_ok=True)

atexit.register(tmpdir.cleanup)

EMAIL_BACKEND = "django.core.mail.outbox"
MAIL_FROM = "orga@orga.org"

COMPRESS_ENABLED = COMPRESS_OFFLINE = False
STATICFILES_STORAGE = "django.contrib.staticfiles.storage.StaticFilesStorage"
GET_SOLO_TEMPLATE_TAG_NAME = "get_solo"

DEBUG = True
DEBUG_PROPAGATE_EXCEPTIONS = True

with suppress(ValueError):
    INSTALLED_APPS.remove("debug_toolbar.apps.DebugToolbarConfig")  # noqa
    MIDDLEWARE.remove("debug_toolbar.middleware.DebugToolbarMiddleware")  # noqa



# Generated by Django 1.11.4 on 2017-08-12 16:05

import byro.common.models.auditable
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [("members", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="MemberProfile",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("member_identifier", models.CharField(max_length=50, null=True)),
                ("birth_date", models.DateField(null=True)),
                ("address", models.CharField(max_length=500, null=True)),
                ("nick", models.CharField(max_length=200, null=True)),
                ("name", models.CharField(max_length=200, null=True)),
                (
                    "member",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="profile_profile",
                        to="members.Member",
                    ),
                ),
            ],
            bases=(byro.common.models.auditable.Auditable, models.Model),
        )
    ]



import urllib

from django.contrib import messages
from django.contrib.auth import authenticate, login, logout
from django.http import HttpRequest, HttpResponseRedirect
from django.shortcuts import redirect
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.timezone import now
from django.utils.translation import gettext as _
from django.views.generic import TemplateView

from byro.common.models import LogEntry


class LoginView(TemplateView):
    template_name = "common/auth/login.html"

    def post(self, request: HttpRequest, *args, **kwargs) -> HttpResponseRedirect:
        username = request.POST.get("username")
        password = request.POST.get("password")
        user = authenticate(username=username, password=password)

        if user is None:
            messages.error(
                request, _("No user account matches the entered credentials.")
            )
            return redirect("common:login")

        if not user.is_active:
            messages.error(request, _("User account is deactivated."))
            LogEntry.objects.create(
                content_object=user,
                user=user,
                action_type="byro.common.login.deactivated",
            )
            return redirect("common:login")

        login(request, user)
        LogEntry.objects.create(
            content_object=user, user=user, action_type="byro.common.login.success"
        )
        url = urllib.parse.unquote(request.GET.get("next", ""))
        if url and url_has_allowed_host_and_scheme(url, request.get_host()):
            return redirect(url)

        return redirect("/")


def logout_view(request: HttpRequest) -> HttpResponseRedirect:
    if request.user:
        LogEntry.objects.create(
            content_object=request.user,
            user=request.user,
            action_type="byro.common.logout",
        )
    logout(request)
    return redirect("/")


class LogInfoView(TemplateView):
    template_name = "common/log/info.html"

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["log_head"] = LogEntry.objects.get_chain_end()
        context["now"] = now()
        return context



# Generated by Django 1.11.13 on 2018-06-26 18:08

from django.db import migrations


def migrate_bookkeeping_model(apps, schema_editor):
    # In the old model, the bank account was implicit. Most VirtualTransaction
    # that have "None" set as source or destination account refer to the bank
    # account here. Except the automatic member fee bookings.
    #
    # Steps:
    #  1. Generate/Find standard accounts: bank, fees_receivable, fees,  donations
    #  2. Find all RealTransaction and process them
    #    a. RealTransaction with VirtualTransaction objects map directly to
    #       new Transaction objects
    #    b. RealTransaction with no VirtualTransaction map to a new unbalanced
    #       Transaction, referencing the bank account
    #  3. Regenerate 'reverses' relation on the new objects
    #  4. Find VirtualTransaction that have not been processed in step 2
    #    a. If they reference a 'member_fees' account, turn them into Transactions
    #       between fees, fees_receivable, and bank, as appropriate
    #    b. Otherwise they probably shouldn't exist, turn them into unbalanced
    #       Transactions
    #
    from byro.bookkeeping.special_accounts import SpecialAccounts

    Account = apps.get_model("bookkeeping", "Account")
    VirtualTransaction = apps.get_model("bookkeeping", "VirtualTransaction")
    RealTransaction = apps.get_model("bookkeeping", "RealTransaction")
    Transaction = apps.get_model("bookkeeping", "Transaction")
    Booking = apps.get_model("bookkeeping", "Booking")

    bank = SpecialAccounts.bank
    fees_receivable = SpecialAccounts.fees_receivable
    fees = SpecialAccounts.fees
    donations = SpecialAccounts.donations

    ACCOUNT_MAP = {"member_donation": donations, "member_fees": fees_receivable}
    map_account = lambda account: ACCOUNT_MAP.get(account.account_category, account)

    handled_vts = set()
    rt_mapping = {}
    for rt in RealTransaction.objects.all():
        t = Transaction.objects.create(
            booking_datetime=rt.booking_datetime, value_datetime=rt.value_datetime
        )
        rt_mapping[rt.pk] = t.pk
        if rt.amount > 0:
            booking = t.bookings.create(debit_account_id=bank.id, amount=rt.amount)
        else:
            booking = t.bookings.create(credit_account_id=bank.id, amount=-rt.amount)
        booking.data = {}
        booking.booking_datetime = rt.booking_datetime
        booking.source = rt.source
        booking.importer = rt.importer
        booking.memo = rt.purpose
        booking.data["other_party"] = rt.originator
        if rt.importer.endswith("csv_importer"):
            booking.data["csv_line"] = rt.data
        elif rt.data:
            booking.data["generic_data"] = rt.data
        booking.save()

        for vt in rt.virtual_transactions.all():
            params = {
                "member": vt.member,
                "amount": vt.amount,
                "data": {"_migration_value_datetime": str(vt.value_datetime)},
            }
            if vt.destination_account:
                t.bookings.create(
                    credit_account_id=map_account(vt.destination_account).id, **params
                )
            if vt.source_account:
                t.bookings.create(
                    debit_account_id=map_account(vt.source_account).id, **params
                )
            handled_vts.add(vt.pk)

    for rt in RealTransaction.objects.filter(reverses__isnull=False).all():
        t = Transaction.objects.get(pk=rt_mapping[rt.pk])
        t.reverses = Transaction.objects.get(pk=rt_mapping[rt.reverses.pk])
        t.save()

    for vt in VirtualTransaction.objects.exclude(pk__in=handled_vts).all():
        t = Transaction.objects.create(value_datetime=vt.value_datetime)

        params = {"member": vt.member, "amount": vt.amount}

        if vt.source_account and vt.destination_account:
            # Special case: Balanced VirtualTransaction
            #  (nothing in the code base generates these, handle them anyway)
            t.bookings.create(
                credit_account_id=map_account(vt.destination_account).id, **params
            )
            t.bookings.create(
                debit_account_id=map_account(vt.source_account).id, **params
            )
        elif vt.source_account and vt.source_account.account_category == "member_fees":
            # Member fee liability
            t.bookings.create(credit_account_id=fees.id, **params)
            t.bookings.create(debit_account_id=fees_receivable.id, **params)

        elif (
            vt.destination_account
            and vt.destination_account.account_category == "member_fees"
        ):
            # Member fee payment, f.e. from make_testdata
            t.bookings.create(credit_account_id=fees_receivable.id, **params)
            t.bookings.create(debit_account_id=bank.id, **params)

        elif vt.source_account:
            # Something else
            t.bookings.create(
                credit_account_id=map_account(vt.source_account).id, **params
            )

        elif vt.destination_account:
            # Something else
            t.bookings.create(
                debit_account_id=map_account(vt.source_account).id, **params
            )

        else:
            # A VirtualTransaction with no source or destination account is a sure sign of multiple exclamation marks
            raise Exception(
                "Migration error: {} without either source or destination account!!!!!".format(
                    vt
                )
            )


class Migration(migrations.Migration):
    dependencies = [
        ("bookkeeping", "0012_auto_20180617_1926"),
        ("common", "0013_logentry_logchain"),
    ]

    operations = [
        migrations.RunPython(migrate_bookkeeping_model, migrations.RunPython.noop)
    ]



# Generated by Django 2.0 on 2017-12-06 19:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("members", "0004_auto_20171013_1436")]

    operations = [
        migrations.AlterField(
            model_name="member",
            name="address",
            field=models.TextField(
                blank=True, max_length=300, null=True, verbose_name="Address"
            ),
        ),
        migrations.AlterField(
            model_name="member",
            name="email",
            field=models.EmailField(
                blank=True, max_length=200, null=True, verbose_name="E-Mail"
            ),
        ),
        migrations.AlterField(
            model_name="member",
            name="name",
            field=models.CharField(
                blank=True, max_length=100, null=True, verbose_name="Name"
            ),
        ),
        migrations.AlterField(
            model_name="member",
            name="number",
            field=models.CharField(
                blank=True,
                max_length=100,
                null=True,
                verbose_name="Membership number/ID",
            ),
        ),
        migrations.AlterField(
            model_name="membership",
            name="amount",
            field=models.DecimalField(
                decimal_places=2,
                help_text="The amount to be paid in the chosen interval",
                max_digits=8,
                verbose_name="membership fee",
            ),
        ),
        migrations.AlterField(
            model_name="membership",
            name="interval",
            field=models.IntegerField(
                choices=[
                    (1, "monthly"),
                    (3, "quarterly"),
                    (6, "biannually"),
                    (12, "annually"),
                ],
                help_text="How often does the member pay their fees?",
                verbose_name="payment interval",
            ),
        ),
    ]



# Generated by Django 2.1.1 on 2018-09-11 19:56

from django.db import migrations


def add_members_to_emails(apps, schema_editor):
    Member = apps.get_model("members", "Member")
    EMail = apps.get_model("mails", "EMail")

    for email in EMail.objects.all():
        for address in email.to.split(","):
            members = Member.objects.filter(email__iexact=address.strip().lower())
            if members.count() == 1:
                email.members.add(members.first())
        email.save()


class Migration(migrations.Migration):
    dependencies = [("mails", "0005_email_members")]

    operations = [
        migrations.RunPython(add_members_to_emails, migrations.RunPython.noop)
    ]



from django.views.generic import TemplateView

from byro.members.models import Member
from byro.members.stats import get_member_statistics


class DashboardView(TemplateView):
    template_name = "office/dashboard.html"

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["member_count"] = Member.objects.all().count()
        context["active_count"] = Member.objects.with_active_membership().count()
        context["stats"] = get_member_statistics()
        return context



# Generated by Django 2.1.1 on 2018-09-29 13:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [("mails", "0007_auto_20180929_0915"), ("common", "0011_logentry")]

    operations = [
        migrations.AddField(
            model_name="configuration",
            name="record_disclosure_template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="mails.MailTemplate",
            ),
        )
    ]



# Generated by Django 2.1.1 on 2018-10-13 16:11

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("documents", "0004_auto_20181013_1611"),
        ("bookkeeping", "0015_auto_20180707_1522"),
    ]

    operations = [
        migrations.CreateModel(
            name="DocumentTransactionLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "document",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="documents.Document",
                    ),
                ),
                (
                    "transaction",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="bookkeeping.Transaction",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="transaction",
            name="documents",
            field=models.ManyToManyField(
                through="bookkeeping.DocumentTransactionLink", to="documents.Document"
            ),
        ),
    ]



BOLD = "\033[1m"
RESET = "\033[0m"


def start_box(size):
    try:
        print("┏" + "━" * size + "┓")
    except (UnicodeDecodeError, UnicodeEncodeError):
        print("-" * (size + 2))


def end_box(size):
    try:
        print("┗" + "━" * size + "┛")
    except (UnicodeDecodeError, UnicodeEncodeError):
        print("-" * (size + 2))


def print_line(string, box=False, bold=False, color=None, size=None):
    text_length = len(string)
    alt_string = string
    if bold:
        string = f"{BOLD}{string}{RESET}"
    if color:
        string = "{color}{string}{RESET}".format(
            color=color, string=string, RESET=RESET
        )
    if box:
        if size:
            if text_length + 2 < size:
                string += " " * (size - text_length - 2)
                alt_string += " " * (size - text_length - 2)
        string = f"┃ {string} ┃"
        alt_string = f"| {string} |"
    try:
        print(string)
    except (UnicodeDecodeError, UnicodeEncodeError):
        try:
            print(alt_string)
        except (UnicodeDecodeError, UnicodeEncodeError):
            print("unprintable setting")



from django.conf import settings
from django.shortcuts import redirect, reverse
from django.urls import resolve
from django.utils import translation

from byro.common.models.configuration import Configuration
from byro.common.signals import unauthenticated_urls


class SettingsMiddleware:
    ALLOWED_URLS = ("settings.registration", "settings.initial", "settings.plugins")

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        url = resolve(request.path_info)
        translation.activate(settings.DEFAULT_LANGUAGE)
        if not request.user.is_anonymous and url.url_name not in self.ALLOWED_URLS:
            config = Configuration.get_solo()
            values = ("name", "backoffice_mail", "mail_from")
            if not all(getattr(config, value, None) for value in values):
                return redirect("office:settings.initial")
        return self.get_response(request)


class PermissionMiddleware:
    UNAUTHENTICATED_URLS = ("login", "logout", "log.info")

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        url = resolve(request.path_info)

        allow = True

        if request.user.is_anonymous and url.url_name not in self.UNAUTHENTICATED_URLS:
            allow = False

        if not allow:
            unauthenticated_urls_matchers = []
            for _receiver, response in unauthenticated_urls.send(self):
                unauthenticated_urls_matchers.extend(response)

            for url_matcher in unauthenticated_urls_matchers:
                if callable(url_matcher):
                    if url_matcher(request, url):
                        allow = True
                        break
                else:
                    if url.view_name == url_matcher:
                        allow = True
                        break

        if not allow:
            return redirect(reverse("common:login") + f"?next={request.path}")
        else:
            return self.get_response(request)



# Generated by Django 2.1.1 on 2018-10-09 18:16

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [("documents", "0002_document_title")]

    operations = [
        migrations.AlterModelOptions(
            name="document", options={"ordering": ("-date", "title", "-id")}
        ),
        migrations.AddField(
            model_name="document",
            name="date",
            field=models.DateField(default=django.utils.timezone.now, null=True),
        ),
        migrations.AddField(
            model_name="document",
            name="content_hash",
            field=models.CharField(max_length=300, null=True),
        ),
        migrations.AlterField(
            model_name="document",
            name="document",
            field=models.FileField(max_length=1000, upload_to="documents/%Y/%m/"),
        ),
    ]



from django.template.defaultfilters import register
from django.utils.translation import gettext_lazy as _

from byro.documents.models import get_document_category_names


@register.filter(name="translate_document_category")
def translate_document_category(data):
    return get_document_category_names().get(data, _("Unknown document type"))



from django.conf.global_settings import LANGUAGES
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from solo.models import SingletonModel

from byro.common.models.choices import Choices
from byro.common.models.log import LogTargetMixin


class ByroConfiguration(LogTargetMixin, SingletonModel):
    """Use this class to build a configuration set that will automatically show
    up on the office settings interface."""

    class Meta:
        abstract = True


class MemberViewLevel(Choices):
    NO = "no"
    NAME_ONLY = "name-only"
    NAME_AND_CONTACT = "name-contact"


class Configuration(ByroConfiguration):
    LOG_TARGET_BASE = "byro.settings"

    name = models.CharField(
        null=True, blank=True, max_length=100, verbose_name=_("Association name")
    )
    address = models.TextField(
        null=True, blank=True, max_length=500, verbose_name=_("Association address")
    )
    url = models.CharField(
        null=True, blank=True, max_length=200, verbose_name=_("Association URL")
    )
    liability_interval = models.IntegerField(
        default=36,
        verbose_name=_("Statute of limitations"),
        help_text=_(
            "For which interval can you make members pay their outstanding fees?"
        ),
    )
    accounting_start = models.DateField(
        null=True,
        blank=True,
        verbose_name=_("Start accounting Membership Fees from"),
        help_text=_(
            "This field is especially useful if the organization was later migrated to byro and the membership fees of members from the past are not to be billed. Leave the field empty if you do not have this requirement and you want to invoice all members from the beginning of their membership."
        ),
    )
    language = models.CharField(
        choices=LANGUAGES,
        null=True,
        blank=True,
        max_length=7,
        verbose_name=_("Language"),
    )
    currency = models.CharField(
        null=True,
        blank=True,
        max_length=3,
        verbose_name=_("Currency"),
        help_text=_("E.g. EUR"),
    )
    currency_symbol = models.CharField(
        default="€",
        max_length=8,
        verbose_name=_("Currency symbol"),
        help_text=_("E.g. €"),
    )
    currency_postfix = models.BooleanField(
        default=True,
        verbose_name=_("Show currency symbol after value"),
        help_text=_(
            "Controls whether the currency symbol comes before or after the monetary value"
        ),
    )
    display_cents = models.BooleanField(
        default=True,
        verbose_name=_("Display cents"),
        help_text=_(
            "When enabled, monetary values include two decimal fractional digits"
        ),
    )
    # Registration form configuration, contains settings for the fields to include when adding a new member
    registration_form = models.JSONField(null=True, blank=True)
    default_order_name = models.CharField(
        max_length=5,
        verbose_name=_("Default sort order"),
        help_text=_(
            "You can always override the automatic sort order if a member has an unusual name."
        ),
        choices=(
            ("first", _("First part of the name")),
            ("last", _("Last part of the name")),
        ),
        default="last",
    )
    default_direct_address_name = models.CharField(
        max_length=5,
        verbose_name=_("Default name when addressing members"),
        help_text=_(
            "This is used in emails, for example. You can always override the automatically chosen name if a member has an unusual name."
        ),
        choices=(
            ("first", _("First part of the name")),
            ("last", _("Last part of the name")),
        ),
        default="first",
    )
    public_base_url = models.URLField(  # Do we want this here or in the settings.py next to SITE_URL?
        max_length=512,
        null=True,
        blank=True,
        verbose_name=_("External base URL of byro installation"),
        help_text=_(
            "This field is used to generate the absolute URL for public pages. Leave it empty if it is the same as this page's base URL."
        ),
    )
    mail_from = models.EmailField(
        null=True,
        blank=True,
        max_length=100,
        verbose_name=_("E-mail address used as sender"),
    )
    backoffice_mail = models.EmailField(
        null=True,
        blank=True,
        max_length=100,
        verbose_name=_("E-mail address for notifications"),
    )
    welcome_member_template = models.ForeignKey(
        to="mails.MailTemplate",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )
    welcome_office_template = models.ForeignKey(
        to="mails.MailTemplate",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )
    leave_member_template = models.ForeignKey(
        to="mails.MailTemplate",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )
    leave_office_template = models.ForeignKey(
        to="mails.MailTemplate",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )
    record_disclosure_template = models.ForeignKey(
        to="mails.MailTemplate",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    form_title = _("General settings")

    def __str__(self):
        return "Settings"

    def get_absolute_url(self):
        return reverse("office:settings.base")



# Generated by Django 1.11.13 on 2018-05-11 16:14

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("members", "0007_member_membership_type")]

    operations = [
        migrations.AddField(
            model_name="member",
            name="member_contact_type",
            field=models.CharField(
                choices=[
                    ("organization", "organization"),
                    ("person", "person"),
                    ("role", "role"),
                ],
                default="person",
                max_length=12,
            ),
        )
    ]



import os

from celery import Celery

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "byro.settings")

from django.conf import settings  # noqa

app = Celery("byro")
app.config_from_object("django.conf:settings", namespace="CELERY")
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)



# Generated by Django 1.11.8 on 2018-03-03 17:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("bookkeeping", "0010_auto_20180113_1500")]

    operations = [
        migrations.AlterField(
            model_name="realtransaction",
            name="importer",
            field=models.CharField(max_length=500, null=True),
        ),
        migrations.AlterField(
            model_name="realtransaction",
            name="originator",
            field=models.CharField(max_length=500),
        ),
    ]



from django.dispatch import receiver
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from byro.common.signals import unauthenticated_urls
from byro.members.models import Member
from byro.members.signals import (
    new_member_mail_information,
    new_member_office_mail_information,
)
from byro.office.signals import member_dashboard_tile, nav_event


@receiver(unauthenticated_urls)
def memberpage_unauthenticated_urls(sender, **kwargs):
    return (lambda a, b: b.view_name.startswith("public:memberpage:"),)


@receiver(nav_event)
def memberpage_primary(sender, **kwargs):
    request = sender
    if request.resolver_match and request.resolver_match.view_name.startswith(
        "public:memberpage"
    ):
        secret_token = request.resolver_match.kwargs.get("secret_token")
        if secret_token:
            kwargs = {"secret_token": secret_token}
            result = [
                {
                    "label": _("Member page"),
                    "url": reverse("public:memberpage:member.dashboard", kwargs=kwargs),
                    "active": request.resolver_match.view_name
                    == "public:memberpage.dashboard",
                }
            ]
            member = Member.all_objects.filter(
                profile_memberpage__secret_token=secret_token
            ).first()
            if member.is_active:
                result.append(
                    {
                        "label": _("Member list"),
                        "url": reverse("public:memberpage:member.list", kwargs=kwargs),
                        "active": request.resolver_match.view_name
                        == "public:memberpage:member.list",
                    }
                )
            return result
    return {}


@receiver(new_member_mail_information)
def new_member_mail_info_memberpage(sender, signal, **kwargs):
    url = sender.profile_memberpage.get_url()
    if url:
        return _("Your personal member page is at {link}").format(
            link=sender.profile_memberpage.get_url()
        )


@receiver(new_member_office_mail_information)
def new_member_office_mail_info_memberpage(sender, signal, **kwargs):
    url = sender.profile_memberpage.get_url()
    if url:
        return _("Their personal member page is at {link}").format(
            link=sender.profile_memberpage.get_url()
        )


@receiver(member_dashboard_tile)
def member_dashboard_page_link(sender, signal, member=None, **kwargs):
    if not member:
        return
    url = member.profile_memberpage.get_url()
    if url:
        return {"url": url, "title": _("Public profile")}



import collections.abc

from django.conf import settings
from django.http import Http404
from django.urls import resolve
from django.utils import formats, translation

from byro.bookkeeping.models import Transaction
from byro.common.models import Configuration, LogEntry
from byro.common.utils import get_version
from byro.mails.models import EMail
from byro.office.signals import nav_event


def byro_information(request):
    ctx = {
        "config": Configuration.get_solo(),
        "pending_mails": EMail.objects.filter(sent__isnull=True).count(),
        "pending_transactions": Transaction.objects.unbalanced_transactions().count(),
        "log_end": LogEntry.objects.get_chain_end(),
        "effective_date_format": formats.get_format(
            "SHORT_DATE_FORMAT", lang=translation.get_language()
        ),
    }

    ctx["effective_date_format_js"] = (
        ctx["effective_date_format"]
        .replace("d", "dd")
        .replace("m", "mm")
        .replace("Y", "yyyy")
    )

    try:
        ctx["url_name"] = resolve(request.path_info).url_name
    except Http404:
        ctx["url_name"] = ""

    if settings.DEBUG:
        ctx["development_warning"] = True

    ctx["byro_version"] = get_version()

    return ctx


def sidebar_information(request):
    _nav_event = []
    for _receiver, response in nav_event.send(request):
        if not response:
            continue
        if isinstance(response, collections.abc.Mapping):
            _nav_event.append(response)
        else:
            _nav_event.extend(response)

    return {"nav_event": _nav_event}



# Generated by Django 1.11.13 on 2018-07-07 14:10

from django.db import models, migrations


def delete_old_accounts(apps, schema_editor):
    Account = apps.get_model("bookkeeping", "Account")

    Account.objects.filter(
        models.Q(account_category="member_donation")
        | models.Q(account_category="member_fees")
    ).delete()


class Migration(migrations.Migration):
    dependencies = [("bookkeeping", "0013_new_data_model")]

    operations = [
        migrations.RemoveField(model_name="realtransaction", name="reverses"),
        migrations.RemoveField(model_name="realtransaction", name="source"),
        migrations.RemoveField(
            model_name="virtualtransaction", name="destination_account"
        ),
        migrations.RemoveField(model_name="virtualtransaction", name="member"),
        migrations.RemoveField(
            model_name="virtualtransaction", name="real_transaction"
        ),
        migrations.RemoveField(model_name="virtualtransaction", name="source_account"),
        migrations.DeleteModel(name="RealTransaction"),
        migrations.DeleteModel(name="VirtualTransaction"),
        migrations.RunPython(delete_old_accounts, migrations.RunPython.noop),
    ]



# Generated by Django 1.11.6 on 2017-10-13 14:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("common", "0002_configuration_mail_from")]

    operations = [
        migrations.AlterField(
            model_name="configuration",
            name="address",
            field=models.TextField(
                blank=True, max_length=500, null=True, verbose_name="address"
            ),
        )
    ]



from django import forms
from django.utils.translation import gettext_lazy as _

from byro.common.models import Configuration
from byro.members.models import Member


class InitialForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self.fields.values():
            field.required = True

    class Meta:
        model = Configuration
        fields = ("name", "backoffice_mail", "mail_from")


class ConfigurationForm(forms.ModelForm):
    update_existing_order_names = forms.BooleanField(
        label=_("Update all order names"),
        initial=False,
        required=False,
    )
    update_existing_direct_address_names = forms.BooleanField(
        label=_("Update all address names"),
        initial=False,
        required=False,
    )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if self.cleaned_data.get("update_existing_order_names"):
            for member in Member.objects.all():
                member.update_order_name(force=True)
        if self.cleaned_data.get("update_existing_direct_address_names"):
            for member in Member.objects.all():
                member.update_direct_address_name(force=True)

    class Meta:
        model = Configuration
        fields = (
            "name",
            "address",
            "url",
            "language",
            "currency",
            "currency_symbol",
            "currency_postfix",
            "display_cents",
            "liability_interval",
            "public_base_url",
            "mail_from",
            "backoffice_mail",
            "accounting_start",
            "default_order_name",
            "update_existing_order_names",
            "default_direct_address_name",
            "update_existing_direct_address_names",
        )



# Generated by Django 1.11.6 on 2017-10-12 18:45

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Configuration",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True, max_length=100, null=True, verbose_name="name"
                    ),
                ),
                (
                    "address",
                    models.CharField(
                        blank=True, max_length=500, null=True, verbose_name="address"
                    ),
                ),
                (
                    "url",
                    models.CharField(
                        blank=True, max_length=200, null=True, verbose_name="url"
                    ),
                ),
                (
                    "language",
                    models.CharField(
                        blank=True, max_length=5, null=True, verbose_name="language"
                    ),
                ),
                (
                    "currency",
                    models.CharField(
                        blank=True, max_length=3, null=True, verbose_name="currency"
                    ),
                ),
                (
                    "registration_form",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True,
                        null=True,
                        verbose_name="registration form configuration",
                    ),
                ),
            ],
            options={"abstract": False},
        )
    ]



import collections.abc
from datetime import timedelta

from django.db.models import Q
from django.http import Http404, HttpResponseRedirect
from django.urls import reverse
from django.utils.timezone import now
from django.views.generic import DetailView, ListView
from django.views.generic.edit import FormMixin

from byro.bookkeeping.models import Booking
from byro.bookkeeping.special_accounts import SpecialAccounts
from byro.common.models.configuration import Configuration, MemberViewLevel
from byro.members.models import Member
from byro.office.signals import member_dashboard_tile
from byro.public.forms import PrivacyConsentForm


class MemberBaseView(DetailView):
    slug_field = "profile_memberpage__secret_token"
    slug_url_kwarg = "secret_token"

    model = Member


class MemberView(FormMixin, MemberBaseView):
    template_name = "public/members/dashboard.html"
    form_class = PrivacyConsentForm

    def get_bookings(self, member):
        account_list = [SpecialAccounts.donations, SpecialAccounts.fees_receivable]
        return (
            Booking.objects.with_transaction_data()
            .filter(
                Q(debit_account__in=account_list) | Q(credit_account__in=account_list),
                member=member,
                transaction__value_datetime__lte=now(),
            )
            .order_by("-transaction__value_datetime")
        )

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        obj = context["member"]
        config = Configuration.get_solo()

        context["config"] = config
        context["bookings"] = self.get_bookings(obj)
        context["member_view_level"] = MemberViewLevel

        _now = now()
        memberships = obj.memberships.order_by("-start").all()
        if not memberships:
            return context

        member_fields = obj.get_fields()
        for field in context["form"]:
            field.meta = (
                member_fields[field.name].getter(obj)
                if field.name in member_fields
                else ""
            ) or ""
        first = memberships[0].start
        delta = timedelta()
        for ms in memberships:
            delta += (ms.end or _now.date()) - ms.start
            if not ms.end or ms.end <= _now.date():
                context["current_membership"] = ms
        context["memberships"] = memberships
        context["member_since"] = {
            "days": int(delta.total_seconds() / (60 * 60 * 24)),
            "years": int(round(delta.days / 365, 1)),
            "first": first,
        }
        context["tiles"] = []
        for __, response in member_dashboard_tile.send(self.request, member=obj):
            if not response:
                continue
            if isinstance(response, collections.abc.Mapping) and response.get(
                "public", False
            ):
                context["tiles"].append(response)
        return context

    def get_form_kwargs(self, *args, **kwargs):
        result = super().get_form_kwargs(*args, **kwargs)
        result["member"] = self.get_object()
        return result

    def get_success_url(self):
        return reverse(
            "public:memberpage:member.dashboard",
            kwargs={"secret_token": self.kwargs["secret_token"]},
        )

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        form = self.get_form()
        if form.is_valid():
            form.save()
        return HttpResponseRedirect(self.get_success_url())


class MemberListView(ListView):
    template_name = "public/members/memberlist.html"
    paginate_by = 50
    context_object_name = "members"

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        config = Configuration.get_solo()
        context["config"] = config
        context["member_view_level"] = MemberViewLevel
        context["member_undisclosed"] = Member.objects.exclude(
            profile_memberpage__is_visible_to_members=True
        ).count()
        return context

    def get_queryset(self):
        secret_token = self.kwargs.get("secret_token")
        if not secret_token:
            raise Http404("Page does not exist")

        member = Member.all_objects.filter(
            profile_memberpage__secret_token=secret_token
        ).first()
        if not member:
            raise Http404("Page does not exist")

        if not member.is_active:
            raise Http404("Page does not exist")

        return Member.objects.filter(profile_memberpage__is_visible_to_members=True)



from django import forms
from django.contrib import messages
from django.shortcuts import redirect
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, FormView, ListView

from byro.bookkeeping.models import RealTransactionSource


class UploadForm(forms.ModelForm):
    class Meta:
        model = RealTransactionSource
        fields = ("source_file",)


class UploadListView(ListView):
    template_name = "office/upload/list.html"
    context_object_name = "uploads"
    model = RealTransactionSource


class CsvUploadView(FormView):
    template_name = "office/upload/add.html"
    model = RealTransactionSource
    form_class = UploadForm

    def form_valid(self, form):
        form.save()
        try:
            form.instance.process()
            messages.success(self.request, _("The upload was processed successfully."))
        except Exception as e:
            messages.error(
                self.request,
                _("The upload was added successfully, but could not be processed: ")
                + str(e),
            )
        self.form = form
        return super().form_valid(form)

    def get_success_url(self):
        return self.request.path


class UploadProcessView(DetailView):
    model = RealTransactionSource

    def post(self, request, *args, **kwargs):
        obj = self.get_object()
        try:
            obj.process()
            messages.success(self.request, _("The upload was processed successfully."))
        except Exception as e:
            messages.error(
                self.request, _("The upload could not be processed: ") + str(e)
            )
        return redirect("office:finance.uploads.list")


class UploadMatchView(DetailView):
    model = RealTransactionSource

    def post(self, request, *args, **kwargs):
        obj = self.get_object()
        errors = success = 0
        for real_transaction in obj.transactions.all():
            try:
                real_transaction.derive_virtual_transactions()
                success += 1
            except Exception:
                errors += 1
        messages.info(
            request,
            "{success} successful matches, {errors} errors.".format(
                success=success, errors=errors
            ),
        )
        return redirect("office:finance.uploads.list")



import django.dispatch

process_transaction = django.dispatch.Signal()
"""
This signal provides a Transaction as sender and expects the receiver
to augment the Transaction with auto-detected information as appropriate.

The common case is a Transaction that is unbalanced and can be augmented
to be a balanced Transaction by adding one or more Bookings.

Recipients MUST NOT change any data in the Transaction or its Bookings if
Transaction.is_read_only is True.
"""

process_csv_upload = django.dispatch.Signal()
"""
This signal provides a RealTransactionSource as sender and expects a list of
one or more Transactions in response.

If the RealTransactionSource has already been processed, no Transactions
should be created, unless you are very sure what you are doing.
"""



# Generated by Django 1.11.8 on 2018-02-24 21:14

from django.db import migrations


def init_templates(apps, schema_editor):
    from byro.mails import default

    MailTemplate = apps.get_model("mails", "MailTemplate")
    Configuration = apps.get_model("common", "Configuration")
    config, _ = Configuration.objects.get_or_create()
    if not config.leave_member_template:
        leave_member = MailTemplate.objects.create(
            subject=default.LEAVE_MEMBER_SUBJECT, text=default.LEAVE_MEMBER_TEXT
        )
        config.leave_member_template = leave_member
    if not config.leave_office_template:
        leave_office = MailTemplate.objects.create(
            subject=default.LEAVE_MEMBER_OFFICE_SUBJECT,
            text=default.LEAVE_MEMBER_OFFICE_TEXT,
        )
        config.leave_office_template = leave_office
    config.save()


class Migration(migrations.Migration):
    dependencies = [("common", "0008_configuration_backoffice_mail")]

    operations = [migrations.RunPython(init_templates, migrations.RunPython.noop)]



# Generated by Django 2.2.12 on 2020-08-09 19:17

from django.db import migrations


def rename(apps, schema_editor):
    if schema_editor.connection.alias != "default":
        return
    Member = apps.get_model("members", "Member")
    for member in Member.objects.all():
        if member.name:
            name_parts = member.name.rsplit(maxsplit=1)
            if len(name_parts) == 1:
                name_parts = name_parts + name_parts
            member.direct_address_name = name_parts[0]
            member.order_name = name_parts[1]
            member.save()


class Migration(migrations.Migration):
    dependencies = [
        ("members", "0011_auto_20200809_1916"),
    ]

    operations = [
        migrations.RunPython(rename, migrations.RunPython.noop),
    ]



from django.template.defaultfilters import register


@register.filter(name="paginate_loop")
def translate_document_category(page_obj, context=5):
    items = [1, page_obj.number, page_obj.paginator.num_pages]
    for i in range(context):
        items.append(max(1, page_obj.number - i))
        items.append(min(page_obj.number + i, page_obj.paginator.num_pages))

    last = 0
    for item in sorted(set(items)):
        if not last + 1 == item:
            yield None
        yield item
        last = item



# Generated by Django 2.1.7 on 2019-02-26 11:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("common", "0013_logentry_logchain")]

    operations = [
        migrations.AddField(
            model_name="configuration",
            name="can_see_other_members",
            field=models.CharField(
                choices=[
                    ("no", "no"),
                    ("name-only", "name-only"),
                    ("name-contact", "name-contact"),
                ],
                default="no",
                max_length=12,
                verbose_name="Members can see other members",
            ),
        ),
        migrations.AddField(
            model_name="configuration",
            name="public_base_url",
            field=models.URLField(
                blank=True,
                help_text="This field is used to generate the absolute URL for public pages. Leave it empty if it is the same as this page's base URL.",
                max_length=512,
                null=True,
                verbose_name="External base URL of byro installation",
            ),
        ),
    ]



# Generated by Django 3.1 on 2020-08-20 20:18

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("common", "0017_auto_20200818_2002"),
    ]

    operations = [
        migrations.AddField(
            model_name="configuration",
            name="default_direct_address_name",
            field=models.CharField(
                choices=[
                    ("first", "First part of the name"),
                    ("last", "Last part of the name"),
                ],
                default="first",
                help_text="This is used in emails, for example. You can always override the automatically chosen name if a member has an unusual name.",
                max_length=5,
                verbose_name="Default name when addressing members",
            ),
        ),
        migrations.AddField(
            model_name="configuration",
            name="default_order_name",
            field=models.CharField(
                choices=[
                    ("first", "First part of the name"),
                    ("last", "Last part of the name"),
                ],
                default="last",
                help_text="You can always override the automatic sort order if a member has an unusual name.",
                max_length=5,
                verbose_name="Default sort order",
            ),
        ),
    ]



from .account import Account, AccountCategory, AccountTag
from .real_transaction import RealTransactionSource
from .transaction import Booking, DocumentTransactionLink, Transaction

__all__ = (
    "Account",
    "AccountTag",
    "AccountCategory",
    "RealTransactionSource",
    "Transaction",
    "Booking",
    "DocumentTransactionLink",
)



# Generated by Django 3.1 on 2020-08-18 20:02

import byro.public.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("public", "0003_auto_20200425_0942"),
    ]

    operations = [
        migrations.AlterField(
            model_name="memberpageprofile",
            name="publication_consent",
            field=models.JSONField(
                blank=True, default=byro.public.models.get_default_consent, null=True
            ),
        ),
    ]



# Generated by Django 2.1.1 on 2018-09-11 19:22

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [("mails", "0003_mailtemplate_reply_to")]

    operations = [
        migrations.AddField(
            model_name="email",
            name="template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="mails.MailTemplate",
            ),
        )
    ]



# Generated by Django 2.1.1 on 2018-10-13 16:11

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("documents", "0003_auto_20181009_1816")]

    operations = [
        migrations.AlterField(
            model_name="document",
            name="direction",
            field=models.CharField(
                choices=[
                    ("incoming", "incoming"),
                    ("outgoing", "outgoing"),
                    ("other", "other"),
                ],
                default="outgoing",
                max_length=8,
            ),
        )
    ]
